/* This file is a component of the SurfAgent Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "gnu_regex.h"

#include "urlcodec.h"

#include "chat.h"	/* for various #defines, get_and_analyze, SITE_MON */

#include "errlog.h"
#include "miscstr.h"

#include "filegard.h"

#include "profile.h"

#include "parsers.h"

#include "csp.h"


char CMOVPERM[] =
  "HTTP/1.0 301 Moved Permanently\n"
  "Location: %s%s%s\n"
  "Content-Type: text/html\n\n"
  "<html>\n"
  "<head><title>301 Moved Permanently</title></head>\n"
  "<body><h1>Moved Permanently</h>\n"
  "The document is now <a href=\"%s%s%s\">here</a>.<p>\n"
  "</body></html>\n"
  ;


/* private type */
struct interceptors {
	char *str;
	char  len;
	char *(*interceptor)( struct http_request *http,
                              struct client_state *csp );
};


/* interceptor functions - used in struct interceptors intercept_patterns */
static char *show_proxy_args(   struct http_request *http,
                                struct client_state *csp );
static char *ij_blocked_url(    struct http_request *http,
                                struct client_state *csp );
static char *sa_feedback_url(   struct http_request *http,
                                struct client_state *csp );
static char *sa_update_categ(   struct http_request *http,
                                struct client_state *csp );
static char *sa_update_monitor( struct http_request *http,
                                struct client_state *csp );


static struct interceptors intercept_patterns[] = {
	{ "show-proxy-args",	15,	show_proxy_args		},
	{ "ij-blocked-url",	14,	ij_blocked_url		},

/* the SurfAgent interception routines... (GLS) */
	{ "sa-feedback-url",	15,	sa_feedback_url		},
	{ "sa-update-categ",	15, 	sa_update_categ		},
	{ "sa-update-monitor",	17, 	sa_update_monitor	},

	{ NULL,			 0,	NULL			}
};



/* dsplit() takes a domain and returns a url_spec
 * structure populated with dbuf, dcnt and dvec.  the other fields
 * in the structure that is returned are zero.
 */
struct url_spec
dsplit( char *domain ) {

  struct url_spec ret;
  char *v[BUFSIZ];
  int size;
  char *p;

  memset( &ret, '\0', sizeof(struct url_spec) );

  if( (p = strrchr(domain, '.')) && (*(++p) == '\0') )
      ret.toplevel = 1;

  ret.dbuf = strdup( domain );

  /* map to lower case */
  for( p = ret.dbuf; *p ; p++ )
    *p = tolower( *p );

  /* split the domain name into components */
  ret.dcnt = ssplit( ret.dbuf, ".", v, SZ(v), 1, 1 );

  if( ret.dcnt <= 0 ) {
    memset( &ret, '\0', sizeof(struct url_spec) );
    return( ret );
  }

  /* save a copy of the pointers in dvec */
  size = ret.dcnt * sizeof( char * );

  if( (ret.dvec = malloc( size )) ) {
    memcpy( ret.dvec, v, size );
  }

  return( ret );

} /* dsplit */

/* the "pattern" is a domain that may contain a '*' as a wildcard.
 * the "fqdn" is the domain name against which the patterns are compared.
 *
 * domaincmp("a.b.c" , "a.b.c")	=> 0 (MATCH)
 * domaincmp("a*.b.c", "a.b.c")	=> 0 (MATCH)
 * domaincmp("b.c"   , "a.b.c")	=> 0 (MATCH)
 * domaincmp(""      , "a.b.c")	=> 0 (MATCH)
 */
int
domaincmp( struct url_spec *pattern, struct url_spec *fqdn ) {

  char **pv, **fv;	/* vectors  */
  int    pn,   fn;	/* counters */
  char  *p,   *f;	/* chars    */
	
  pv = pattern->dvec;
  pn = pattern->dcnt;

  fv = fqdn->dvec;
  fn = fqdn->dcnt;

  while( (pn > 0) && (fn > 0) ) {

    p = pv[--pn];
    f = fv[--fn];

    for( ; *p && *f && (*p == tolower(*f)); p++, f++ );

    if( (*p != tolower(*f)) && (*p != '*') )
      return( 1 );
  }

  if( pn > 0 )
    return( 1 );

  return(0);
} /* domaincmp */



/******************************************************************************/

/* the following functions are invoked from chat() to catch url's that are
   either fake and thus need to be intercepted by JunkBuster, or there's
   something else wrong with them (i.e., on the blocklist, etc.)
 */


/* checks if the URL matches something on the blocklist */
char *
block_url( struct http_request *http ) {

  char *templ_str =
    "HTTP/1.0 202 Request for blocked URL\n"
    "Pragma: no-cache\n"
    "Last-Modified: Thu Jul 31, 1997 07:42:22 pm GMT\n"
    "Expires:       Thu Jul 31, 1997 07:42:22 pm GMT\n"
    "Content-Type: text/html\n\n"
    "<html>\n"
    "<head>\n"
    "<title>Internet Junkbuster: Request for blocked URL</title>\n"
    "</head>\n"
    BODY
    "<center><h1>" BANNER "</h1></center>\n"
    BANNER " is blocking this page because it "
    "<a href=http://internet.junkbuster.com/ij-blocked-url?%s+%s+%s>matches</a>"
    " one of the patterns in the blocklist.\n"
    "<p>Click <a href=\"%s%s\">here</a> to circumvent the block and view the "
    "page anyway (may require more than one click).\n"
    "</body>\n"
    "</html>\n"
    ;

  char *bspec;
  char *p;
  char *hostport, *path, *spec;
  int n;

  if( (bspec = is_blocked_url(http)) ) {

    hostport = url_encode( http->hostport );
    path     = url_encode( http->path );
    spec     = url_encode( bspec );

    n  = strlen( templ_str );
    n += strlen( hostport );
    n += strlen( path );
    n += strlen( spec );
    n += strlen( http->path );
    n += strlen( passageword );

    p = malloc( n );

    sprintf( p, templ_str, hostport, path, spec, http->path, passageword );

    /* objects were allocated by url_encode() */
    freez( hostport );
    freez( path );
    freez( spec );

    /* object was allocated by is_blocked_url() */
    free( bspec );	/* we know for sure it's NOT NULL */

    return( p );
  }

  return( NULL );

} /* block-url */


/* this checks if the url is fake and needs to be intercepted by JunkBuster */
/* intercept_url() checks the URL `basename' against a list of URLs
 * to snarf.  If it matches, it calls the associated function which
 * returns an HTML page to send back to the client.
 */
char *
intercept_url( struct client_state *csp ) {

  struct http_request *http;
  char *basename;
  struct interceptors *v;

  http = csp->http;
  basename = strrchr( http->path, '/' );

  if( !basename )
    return( NULL );

  ++basename;	/* first char past the last slash */

  if( *basename != '\0' )
    for( v = intercept_patterns; v->str; v++ )
      if( strncmp(basename, v->str, v->len) == 0 )
        return( (v->interceptor)(http, csp) );

  return( NULL );

} /* intercept_url */

/******************************************************************************/

/* interceptor functions - used in struct interceptors intercept_patterns */


/* display the options used when JunkBuster was started */
static char *
show_proxy_args( struct http_request *http, struct client_state *csp ) {

  /* forward the call to the routine in the filegard module */
  return( get_proxy_args() );

} /* show_proxy_args */


/* This function offers an explanation as to why a site was blocked
 * (when the user clicks the link offerred on the block notification)
 */
static char *
ij_blocked_url( struct http_request *http, struct client_state *csp ) {

  char *templ_str =
    "HTTP/1.0 200 OK\r\n"
    "Pragma: no-cache\n"
    "Last-Modified: Thu Jul 31, 1997 07:42:22 pm GMT\n"
    "Expires:       Thu Jul 31, 1997 07:42:22 pm GMT\n"
    "Content-Type: text/html\n\n"
    "<html>\n"
    "<head>\n"
    "<title>Internet Junkbuster: Request for blocked URL</title>\n"
    "</head>\n"
    BODY
    "<center><h1>" BANNER "</h1></center>\n"
    "The " BANNER " Proxy "
    "<A href=\"http://www.junkbuster.com\"> (http://www.junkbuster.com) </A> "
    "intercepted the request for %s%s because the URL matches the following "
    "pattern in the blockfile: %s\n"
    "</body>\n"
    "</html>\n"
    ;

  int n;
  char *hostport, *path, *pattern, *p, *v[9];

  if( (n = ssplit(http->path, "?+", v, SZ(v), 1, 1)) == 4 ) {
    hostport = url_decode( v[1] );
    path     = url_decode( v[2] );
    pattern  = url_decode( v[3] );
  } else {
    hostport = strdup( "undefined_host" );
    path     = strdup( "/undefined_path" );
    pattern  = strdup( "undefined_pattern" );
  }

  n  = strlen(templ_str );
  n += strlen(hostport  );
  n += strlen(path      );
  n += strlen(pattern   );

  if( (p = malloc(n)) )
    sprintf( p, templ_str, hostport, path, pattern );

  freez( hostport );
  freez( path     );
  freez( pattern  );

  return( p );

} /* ij_blocked_url */


/* This function offers the feedback menu in response to clicking on the
 * |Feedback| button...
 */
static char *
sa_feedback_url( struct http_request *http, struct client_state *csp ) {

  char *templ_str =
    "HTTP/1.0 200 OK\r\n"
    "Pragma: no-cache\n"
    "Last-Modified: Mon Nov 09, 1998 67:16:10 pm GMT\n"
    "Expires:       Mon Nov 09, 1998 67:16:10 pm GMT\n"
    "Content-Type: text/html\n\n"
    "<html>\n"
    "<head>\n"
    "<title>SurfAgent: Request feedback on URL</title>\n"
    "</head>\n"
    BODY
    "<center><h1>" BANNER "</h1></center>\n"
    BANNER " intercepted the request for feedback on the document"
    "<center><font color=\"yellow\">%s%s</font></center></p>\n"
    "<hr>\n"
    "<h2>User Profile:</h2>\n"
    "%s"	/* Existing category selector goes here... */
    "<form action=\"http://%s/sa-update-categ\" method=\"GET\">\n"
    "<input type=\"hidden\" name=\"path\" value=\"%s\">\n"
    "Create a new user profile category with the document:<br>\n New Category: "
    "<input type=\"text\" name=\"newcateg\" size=\"10\">\n"
    "<input type=\"submit\" value=\"Create!\">\n"
    "</form>\n"
    "<hr>\n"
    "<h2>Site Monitor:</h2>"
    "<form action=\"http://%s/sa-update-monitor\" method=\"GET\">\n"
    "<input type=\"hidden\" name=\"path\" value=\"%s\">\n"
    "Add the document to the list of monitored sites: "
    "<input type=\"hidden\" name=\"action\" value=\"add\">\n"
    "<input type=\"submit\" value=\"Add!\">\n"
    "</form>"
    "%s"	/* Selector for removal of monitored sites goes here */
    "%s"	/* table containing site-monitor results goes here */
    "<hr>\n"
    "</body>\n"
    "</html>\n"
    ;

  char *templ_profile_selector =
    "<br><form action=\"http://%s/sa-update-categ\" method=\"GET\">\n"
    "<input type=\"hidden\" name=\"path\" value=\"%s\">\n"
    "Add the document to an existing user profile category:<br>\n"
    "<select name=\"oldcateg\">"
    "%s"	/* Catlist goes here: e.g., "\n<option> CategX" */
    "\n</select>\n"
    "<input type=\"submit\" value=\"Add!\">\n"
    "</form>\n"
    ;

  char *templ_monitored_selector =
    "<form action=\"http://%s/sa-update-monitor\" method=\"GET\">\n"
    "<input type=\"hidden\" name=\"path\" value=\"%s\">\n"
    "<input type=\"hidden\" name=\"action\" value=\"del\">\n"
    "Remove currently monitored sites from the list:<br>\n"
    "<select multiple size=\"3\" name=\"remove\">"
    "%s"	/* Monlist goes here: e.g., "\n<option> www.foo.gov/news" */
    "\n</select>\n"
    "<input type=\"submit\" value=\"Remove!\">\n"
    "</form><br>\n"
    ;

  char *templ_mon_res_header =
    "The following table contains links to articles retrieved from the "
    "pages monitored by SiteMonitor. The (few) ones with poor scores have "
    "been retrieved on purpose, with a small probability, in order to offer "
    "the possibility of exploration...<br>\n"
    ;

  char *templ_str_err =
    "HTTP/1.0 200 OK\r\n"
    "Pragma: no-cache\n"
    "Last-Modified: Mon Nov 09, 1998 67:16:10 pm GMT\n"
    "Expires:       Mon Nov 09, 1998 67:16:10 pm GMT\n"
    "Content-Type: text/html\n\n"
    "<html>\n"
    "<head>\n"
    "<title>SurfAgent: Feedback Error</title>\n"
    "</head>\n"
    BODY
    "<center><h1>" BANNER "</h1></center>\n"
    BANNER " Could not understand your request...<p>\n"
    "You should only arrive here by clicking the \"Feedback\" button at the "
    "end of downloaded documents...\n"
    "</body>\n"
    "</html>\n"
    ;

  int n;
  char *p, *v[4], *hostport, *path;
  char *catlist, *catsel, *monlist, *monsel, *monres, *monrestable;

  if( (n = ssplit(http->path, "?!", v, SZ(v), 1, 1)) == 4 ) {
    hostport = url_decode( v[1] );
    path     = url_decode( v[2] );
    catlist  = url_decode( v[3] );
  } else {
    return( strdup(templ_str_err) );
  }

  /* build monsel of existing monitored sites, if any */
  monsel = NULL;
  if( (monlist = get_monlist(csp->user_id)) ) {
    n  = strlen( templ_monitored_selector );
    n += strlen( hostport );
    n += strlen( path );
    n += strlen( monlist );

    if( !(monsel = malloc(n)) ) {
      errlogf( ERN, "sa_feedback_url(): malloc monsel failed !!\n" );
      exit( 1 );
    }
    sprintf( monsel, templ_monitored_selector, hostport, path, monlist );
    free( monlist );
  }

  /* build catsel of existing categories, if applicable */
  catsel = NULL;
  if( strcmp(catlist, "<sa_none>") != 0 ) {
    n  = strlen( templ_profile_selector );
    n += strlen( hostport );
    n += strlen( path );
    n += strlen( catlist );

    if( !(catsel = malloc(n)) ) {
      errlogf( ERN, "sa_feedback_url(): malloc catsel failed !!\n" );
      exit( 1 );
    }
    sprintf( catsel, templ_profile_selector, hostport, path, catlist );
  }

  /* get monres if available */
  monrestable = NULL;
  if( (monres = get_monres(csp->user_id)) ) {
    monrestable = strdup( templ_mon_res_header );
    monrestable = strsav( monrestable, monres );
  }

  n  = strlen( templ_str );
  n += strlen( hostport  ) * 3;
  n += strlen( path      ) * 3;
  if( monsel )
    n += strlen( monsel );
  if( monrestable )
    n += strlen( monrestable );
  if( catsel )
    n += strlen( catsel );

  if( (p = malloc(n)) )
    sprintf( p, templ_str,
                hostport, path, (catsel)? catsel : "", hostport, path,
                hostport, path, (monsel)? monsel : "",
                (monrestable)? monrestable : "" );

  freez( monrestable );
  freez( monres );
  freez( monsel );
  freez( catsel );

  freez( hostport );
  freez( path );
  freez( catlist );

  return( p );

} /* sa_feedback_url */


/* Intercept request to update category. Request a download_and_analyze() to
   get the wordvector, then request an update_profile() to make the necessary
   changes to the user profile. Finally, send back a CMOVPERM to have the
   client reload the page after everything else is taken care of...
 */
static char *
sa_update_categ( struct http_request *http, struct client_state *csp ) {

  char *templ_str_err =
    "HTTP/1.0 200 OK\r\n"
    "Pragma: no-cache\n"
    "Last-Modified: Mon Nov 09, 1998 67:16:10 pm GMT\n"
    "Expires:       Mon Nov 09, 1998 67:16:10 pm GMT\n"
    "Content-Type: text/html\n\n"
    "<html>\n"
    "<head>\n"
    "<title>SurfAgent: User Profile Update Error</title>\n"
    "</head>\n"
    BODY
    "<center><h1>" BANNER "</h1></center>\n"
    BANNER " Could not understand your request...<p>\n"
    "You should only arrive here by requesting an update to your "
    "user profile, case in which you must have omitted some information...\n"
    "<p>Also keep in mind that neither of your categories can be named "
    "<b><i>" SITE_MON "</i></b>, which is a reserved name for the "
    "Site Monitor service...\n"
    "</body>\n"
    "</html>\n"
    ;

  int n;
  char buf[BUFSIZ], *v[9], *path, *categ_type, *categ, *p;

  if( (n = ssplit(http->path, "?&=", v, SZ(v), 1, 1)) == 5 ) {
    path       = url_decode( v[2] );
    categ_type = url_decode( v[3] );
    categ      = url_decode( v[4] );
  } else {
   return( strdup(templ_str_err) );
  }

  /* It's real important to avoid adding the score value to the category
   * name. Therefore, we need to skip the score. Currently the score size
   * is 6 (four characters for the value, plus two parantheses). See profile.c
   * function update_profile() for more details...
   */
  if( strcmp(categ_type, "newcateg") == 0 )
    p = categ;
  else
    p = categ + 6;

  /* make sure the category doesn't have the same name as the SITE_MON file */
  if( strcmp(p, SITE_MON) == 0 ) {
    freez( categ );
    freez( categ_type );
    freez( path );
    return( strdup(templ_str_err) );
  }

  /* make the correct url */
  sprintf( buf, "%s%s%s", http->protocol, http->hostport, path );

  /* get word vector for target document, place it in csp */
  /* note: the csp cleanup routine will de-allocate this wordvec */
  get_and_analyze( buf, NULL, &(csp->wordvec) );

  /* now update the user profile (p is the category name) */
  update_profile( csp->wordvec, csp->user_id, p );

  /* create "moved permanently" page to return */
  sprintf( buf, CMOVPERM, http->protocol, http->hostport, path,
                          http->protocol, http->hostport, path );

  /* clean up */
  freez( categ );
  freez( categ_type );
  freez( path );

  return( strdup(buf) );

} /* sa_update_categ */


/*---------------------------------------------------------------------------*/
/* intercept request to update the site monitor.                             */
/* Via URL we are given the hostport and path of the site we started from    */
/*  (by clicking feedback). If no further parameters are given, we add this  */
/*  site to the user's monitored list. If there are more parameters, we must */
/*  remove all given sites from the list (and we add nothing in this case)   */
/*---------------------------------------------------------------------------*/
static char *
sa_update_monitor( struct http_request *http, struct client_state *csp ) {

  char *template_path_err =
    "HTTP/1.0 200 OK\r\n"
    "Pragma: no-cache\n"
    "Last-Modified: Mon Nov 09, 1998 67:16:10 pm GMT\n"
    "Expires:       Mon Nov 09, 1998 67:16:10 pm GMT\n"
    "Content-Type: text/html\n\n"
    "<html>\n"
    "<head>\n"
    "<title>SurfAgent: Site Monitor Update Error</title>\n"
    "</head>\n"
    BODY
    "<center><h1>" BANNER "</h1></center>\n"
    BANNER " Could not understand your request...<p>\n"
    "You should only arrive here by requesting an update to your site monitor, "
    "case in which you must have omitted some information...<br>\n"
    "If you KNOW you did nothing wrong, probably an error occurred in "
    "sa_update_monitor(): Most likely, the v array was "
    "too small for ssplit().<br>\n"
    "If this makes no sense to you, please contact your SurfAgent admin...\n"
    "</body>\n"
    "</html>\n"
    ;

  int n, i;
  char buf[BUFSIZ], *path, *action, *arg, *v[20];

  if( (n = ssplit(http->path, "?&=", v, SZ(v), 1, 1)) < 5 ) {
   return( strdup(template_path_err) );
  } else {
    path   = url_decode( v[2] );
    action = url_decode( v[4] );	/* "add" or "del" */
  }

  if( strcmp(action, "add") == 0 ) {
    sprintf( buf, "%s%s%s", http->protocol, http->hostport, path );
    if( add_monlist(csp->user_id, buf) != 0 )
      errlogf( ERR, "sa_update_monitor(): error adding site: %s\n", buf );
  }

  if( strcmp(action, "del") == 0 )
    for( i = 6; i < n; i += 2 ) {
      arg = url_decode( v[i] );
      if( del_monlist(csp->user_id, arg) != 0 )
        errlogf( ERR, "sa_update_monitor(): error deleting site: %s\n", arg );
      freez( arg );
    }

  sprintf( buf, CMOVPERM, http->protocol, http->hostport, path,
                          http->protocol, http->hostport, path );

  freez( path );
  freez( action );

  return( strdup(buf) );

} /* sa_update_monitor */
