/* This file is a component of the SurfAgent Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/**********************************  hash.c  ***********************************

    Purpose: Offer services related to a hash table data structure (insertion
             and lookup of strings, creation and deletion of a hash table

    Notes:   none.
**/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>
#include "hash.h"
#include "errlog.h"
#include "miscstr.h"


#define START_CAPACITY 2048		/* init. capacity for new hash table */
#define MAX_COLLISIONS 500		/* allowed collisions before expand */
#define HASH_CONSTANT  507		/* used in key generation */

/* string and count pair structure */
typedef struct {
  char *word;
  int count;
} StringCount;

/* hash table structure */
typedef struct _HashTable {
  int capacity;
  int nstrings;
  int collisions;
  StringCount *strings;
} HashTableStruct;


/*============================================================================*/
/* hash function: convert a string str into an integer key for the hash table */
/*============================================================================*/
static int
HashKey( const char *str, HashTable ht ) {

  int i, key = 0;

  for( i = 0; i < strlen(str); ++i )
    key = ( (int)str[i] + HASH_CONSTANT * key );

  if( key < 0 )
    key = -key;

  return( key % ht->capacity );

} /* HashKey */


/*============================================================================*/
/* simple insertion function - does not expand the table, but sets collisions */
/* returns new count on success, -1 if out of capacity                        */
/*============================================================================*/
static int
HashSimpleIns( HashTable ht, const char *str, int flg, int cnt ) {

  int key, retval;

  if( !ht || !str || (cnt < 0) || (ht->nstrings >= ht->capacity) ) {
    errlogf( ERR, "HashSimpleIns(): invalid ht, str, or cnt or cap. reached"
                  "Should never happen\n" );
    exit( 1 );
  }

  for( key = HashKey(str, ht), ht->collisions = 0;
       ht->strings[key].word && strcmp(str, ht->strings[key].word) != 0;
       key = (key + 1) % ht->capacity, ++(ht->collisions) );

  if( !(ht->strings[key].word) ) {
    ht->strings[key].word = strdup( str );
    ++(ht->nstrings);
  }

  retval = ht->strings[key].count;
  switch( flg ) {
    case HASH_INS_INC:
      retval += cnt;
      break;
    case HASH_INS_SET:
      retval = cnt;
      break;
    default:
      errlogf( ERR, "HashSimpleIns(): bad flg=%d\n", flg );
      exit( 1 );
  }
  ht->strings[key].count = retval;

  return( retval );

} /* HashSimpleIns */


/*================================================*/
/* expand the hash table by doubling the capacity */
/*================================================*/
static void
HashExpandTable( HashTable ht ) {

  StringCount *old_strings;
  int i, old_capacity, old_nstrings;

  if( ht->capacity >= (MAXINT / 2) ) {
    errlogf( ERR, "HashExpandTable(): failed to double capacity\n" );
    exit( 1 );
  }

  old_strings = ht->strings;
  old_capacity = ht->capacity;
  old_nstrings = ht->nstrings;

  ht->capacity *= 2;
  ht->nstrings = 0;
  if( !(ht->strings = zalloc(ht->capacity * sizeof(StringCount))) ) {
    errlogf( ERN, "HashExpandTable(): failed to allocate new table\n" );
    exit( 1 );
  }

  for( i = 0; i < old_capacity; ++i )
    if( old_strings[i].word )
      HashSimpleIns(ht, old_strings[i].word,
                    HASH_INS_SET, old_strings[i].count);

  if( old_nstrings != ht->nstrings ) {
    errlogf( ERR, "HashExpandTable(): old(%d) != new(%d) nstrings mismatch\n",
                  old_nstrings, ht->nstrings );
    exit( 1 );
  }

  free( old_strings );

} /* HashExpandTable */


/*=========================*/
/* Create empty hash table */
/*=========================*/
HashTable
CreateHashTable( void ) {

  HashTable ht;

  if( !(ht = zalloc(sizeof(HashTableStruct))) ||
      !(ht->strings = zalloc(START_CAPACITY * sizeof(StringCount))) ) {
    errlogf( ERN, "CreateHashTable(): failed to allocate new table\n" );
    return( NULL );
  }

  ht->capacity = START_CAPACITY;
  ht->nstrings = 0;
  ht->collisions = 0;

  return( ht );

} /* CreateHashTable */


/*=======================================================================*/
/* Insert a string into the hash table;                                  */
/*  flg controls whether the count is set to the parameter cnt or simply */
/*      incremented by 1 (in which case cnt is ignored)                  */
/*  the return value is the new count (>0) on success, and -1 on failure */
/*=======================================================================*/
int
HashInsStr( HashTable ht, const char *str, int flg, int cnt ) {

  /* parameter sanity check */
  if( !ht || !str || (cnt < 0) ) {
    errlogf( ERR, "HashInsStr(): invalid parameter (ht=%p, str=%p, cnt=%d)\n",
                  ht, str, cnt );
    return( -1 );
  }

/* This is debugging stuff for the growth of the hash table ######
if( ht->nstrings >= ht->capacity )
  errlogf( ERR, "blah: capacity (cap%d,num%d,col%d)\n", ht->capacity, ht->nstrings, ht->collisions );
if( ht->collisions >= MAX_COLLISIONS )
  errlogf( ERR, "blah: collisions (cap%d,num%d,col%d)\n", ht->capacity, ht->nstrings, ht->collisions );
##### */

  if( (ht->nstrings >= ht->capacity) || (ht->collisions >= MAX_COLLISIONS) )
    HashExpandTable( ht );

  return( HashSimpleIns(ht, str, flg, cnt) );

} /* HashInsStr */


/*======================================================*/
/* Load a hash table from a file.                       */
/*======================================================*/
HashTable
HashLoad( char *filename, HashTable ht ) {

  FILE *fp;
  char buf[BUFSIZ];
  int cnt;

  /* create empty table */
  if( !ht && !(ht = CreateHashTable()) ) {
    errlogf( ERR, "HashLoad(): error creating new hash table.\n" );
    exit( 1 );
  }

  /* open source file */
  if( !(fp=fopen(filename, "r")) ) {
    errlogf( ERN, "HashLoad(): error opening word vector file %s\n",
                  filename );
    return( ht );
  }

  /* read the file: each line is a pair of count and word */
  while( fscanf(fp, "%d%s", &cnt, buf) == 2 ) {
    buf[BUFSIZ-1] = '\0';	/* make sure string is terminated */
    if( HashInsStr(ht, buf, HASH_INS_SET, cnt) < 0 ) {
      errlogf( ERR, "HashLoad(): Error adding string %s from file %s\n",
                    buf, filename );
      exit( 1 );
    }
  }

  /* close the input file when done */
  fclose( fp );

  /* return the resulting hash table */
  return( ht );

} /* HashLoad */


/*==================================================================*/
/* Save a hash table to a file. Returns 0 on success, -1 on failure */
/*==================================================================*/
int
HashSave( HashTable ht, char *filename ) {

  FILE *fp;
  int i;

  /* open destination file */
  if( !(fp=fopen(filename, "w")) ) {
    errlogf( ERN, "TrieSave(): error writing trie to file %s\n", filename );
    return( -1 );
  }

  /* write all the strings and their counts to the file */
  for( i = 0; i < ht->capacity; ++i )
    if( ht->strings[i].word )
      fprintf( fp, "%20d %s\n", ht->strings[i].count, ht->strings[i].word );

  /* close the destination file when done writing */
  fclose( fp );

  /* return success */
  return( 0 );

} /* HashSave */


/*=====================*/
/* Delete a hash table */
/*=====================*/
HashTable
DeleteHashTable( HashTable ht ) {

  if( ht ) {
    if( ht->strings )
      free( ht->strings );
    free( ht );
  }

  return( NULL );

} /* DeleteHashTable */
