/* This file is a component of the SurfAgent Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/********************************* htmparse.c **********************************

     Purpose: HTML parsing module: scan the html text, separating it
              out into words, and figuring out when we're in a link and what
              the URL of the link is...
     Notes: This module is implemented as an automaton: there are states between
            which transitions occur based on characters; on each transition
            there is a possible action associated with it.

     Provenience: Written by Gabriel L. Somlo, 5/4/1999

**/

/*================================ Local Stuff ===============================*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "errlog.h"

#include "htmparse.h"
#include "wordvec.h"		/* for the use of the WordVec type */
#include "linkvec.h"		/* for the use of the LinkVec type */
#include "stopword.h"		/* for NCHARS */


#define NSTATES 25

#define HAVE_URL  1		/* a url got stored in the outbuf */
#define HAVE_WORD 2		/* a word got stored in the outbuf */
#define HAVE_BASE 3		/* a base url got stored in the outbuf */


typedef struct _HtmlParser {
  int crtstate;				/* current state */
  char outbuf[BUFSIZ];			/* word/url buffer */
  int outchr;				/* number of chars in buffer */
  TrieNode stopw;			/* state for stopword remover */
  int in_link;				/* bool: we're going through a link */
  int outcome;				/* result of last action */
} HtmlParser;

typedef void (*parsefuncptr)( HPP, char );


static struct {
  int nextstate;
  parsefuncptr action;
} cell[NSTATES][NCHARS];		/* the automaton */


/*----------------------------------------------------------*/
/* store a character into the outbuf - convert to lowercase */
/*----------------------------------------------------------*/
static void
store( HPP hpp, char c ) {

  /* if outbuf is full, crunch it */
  if( hpp->outchr >= (BUFSIZ - 1) )
    hpp->outchr = 0;

  hpp->outbuf[(hpp->outchr)++] = tolower( c );

  hpp->stopw = StopNextChar( hpp->stopw, tolower(c) );

  hpp->outcome = 0;

} /* store */


/*---------------------------------------------------*/
/* store a character into the outbuf - preserve case */
/*---------------------------------------------------*/
static void
store_url( HPP hpp, char c ) {

  /* if outbuf is full, crunch it */
  if( hpp->outchr >= (BUFSIZ - 1) )
    hpp->outchr = 0;

  hpp->outbuf[(hpp->outchr)++] = c;

  hpp->outcome = 0;

} /* store_url */


/*------------------*/
/* reset the outbuf */
/*------------------*/
static void
reset( HPP hpp, char c ) {

  hpp->outchr = 0;

  hpp->stopw = StopReset();

  hpp->outcome = 0;

} /* reset */


/*----------------------------------------*/
/* deal with a finished URL in the outbuf */
/*----------------------------------------*/
static void
have_url( HPP hpp, char c ) {

  hpp->outbuf[hpp->outchr] = '\0';

  hpp->outcome = HAVE_URL;

} /* have_url */


/*-----------------------------------------*/
/* deal with a finished word in the outbuf */
/*-----------------------------------------*/
static void
have_word( HPP hpp, char c ) {

  hpp->outbuf[hpp->outchr] = '\0';

  hpp->outcome = HAVE_WORD;

} /* have_word */


/*---------------------------------------------*/
/* deal with a finished base URL in the outbuf */
/*---------------------------------------------*/
static void
have_base( HPP hpp, char c ) {

  hpp->outbuf[hpp->outchr] = '\0';

  hpp->outcome = HAVE_BASE;

} /* have_url */


/*--------------------------------------*/
/* signal the start of analyzing a link */
/*--------------------------------------*/
static void
start_link( HPP hpp, char c ) {

  hpp->in_link = 1;

  hpp->outcome = 0;

} /* start_link */


/*--------------------------------------*/
/* signal the end of analyzing a link */
/*--------------------------------------*/
static void
end_link( HPP hpp, char c ) {

  hpp->in_link = 0;

  hpp->outcome = 0;

} /* end_link */


/*------------------------------------------------------------*/
/* initialize the state-transition table for the html parser. */
/*------------------------------------------------------------*/
static void
InitHtmlParseTable( void ) {

  int state, chr;

  state = 0;
  for( chr = 0; chr < NCHARS; ++chr )
    if( isalpha(chr) ) {
      cell[state][chr].nextstate = 13;
      cell[state][chr].action = store;
    } else {
      cell[state][chr].nextstate = 0;
      cell[state][chr].action = NULL;
    }
  cell[state]['<'].nextstate = 1;

  state = 1;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].action = NULL;
    if( isalpha(chr) )
      cell[state][chr].nextstate = 12;
    else
      cell[state][chr].nextstate = 0;
  }
  cell[state]['a'].nextstate = 2;
  cell[state]['A'].nextstate = 2;
  cell[state]['/'].nextstate = 10;
  cell[state]['b'].nextstate = 14;
  cell[state]['B'].nextstate = 14;

  state = 2;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].action = NULL;
    if( isspace(chr) )
      cell[state][chr].nextstate = 2;
    else
      cell[state][chr].nextstate = 12;
  }
  cell[state]['h'].nextstate = 3;
  cell[state]['H'].nextstate = 3;
  cell[state]['>'].nextstate = 0;

  state = 3;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].action = NULL;
    cell[state][chr].nextstate = 12;
  }
  cell[state]['r'].nextstate = 4;
  cell[state]['R'].nextstate = 4;
  cell[state]['>'].nextstate = 0;

  state = 4;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].action = NULL;
    cell[state][chr].nextstate = 12;
  }
  cell[state]['e'].nextstate = 5;
  cell[state]['E'].nextstate = 5;
  cell[state]['>'].nextstate = 0;

  state = 5;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].action = NULL;
    cell[state][chr].nextstate = 12;
  }
  cell[state]['f'].nextstate = 6;
  cell[state]['F'].nextstate = 6;
  cell[state]['>'].nextstate = 0;

  state = 6;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].action = NULL;
    if( isspace(chr) )
      cell[state][chr].nextstate = 6;
    else
      cell[state][chr].nextstate = 12;
  }
  cell[state]['='].nextstate = 7;
  cell[state]['>'].nextstate = 0;

  state = 7;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].action = NULL;
    if( isspace(chr) )
      cell[state][chr].nextstate = 7;
    else
      cell[state][chr].nextstate = 12;
  }
  cell[state]['\"'].nextstate = 8;
  cell[state]['>'].nextstate = 0;

  state = 8;
  for( chr = 0; chr < NCHARS; ++chr )
    if( isspace(chr) ) {
      cell[state][chr].nextstate = 12;
      cell[state][chr].action = reset;
    } else {
      cell[state][chr].nextstate = 8;
      cell[state][chr].action = store_url;
    }
  cell[state]['\"'].nextstate = 9;
  cell[state]['\"'].action = have_url;
  cell[state]['>'].nextstate = 0;
  cell[state]['>'].action = reset;

  state = 9;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].nextstate = 9;
    cell[state][chr].action = NULL;
  }
  cell[state]['>'].nextstate = 0;
  cell[state]['>'].action = start_link;

  state = 10;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].nextstate = 12;
    cell[state][chr].action = NULL;
  }
  cell[state]['a'].nextstate = 11;
  cell[state]['A'].nextstate = 11;
  cell[state]['>'].nextstate = 0;

  state = 11;
  for( chr = 0; chr < NCHARS; ++chr )
    if( isspace(chr) ) {
      cell[state][chr].nextstate = 12;
      cell[state][chr].action = end_link;
    } else {
      cell[state][chr].nextstate = 0;
      cell[state][chr].action = NULL;
    }
  cell[state]['>'].action = end_link;

  state = 12;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].nextstate = 12;
    cell[state][chr].action = NULL;
  }
  cell[state]['>'].nextstate = 0;

  state = 13;
  for( chr = 0; chr < NCHARS; ++chr )
    if( isalnum(chr) ) {
      cell[state][chr].nextstate = 13;
      cell[state][chr].action = store;
    } else {
      cell[state][chr].nextstate = 0;
      cell[state][chr].action = have_word;
    }
  cell[state]['<'].nextstate = 1;

  state = 14;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].nextstate = 12;
    cell[state][chr].action = NULL;
  }
  cell[state]['a'].nextstate = 15;
  cell[state]['A'].nextstate = 15;
  cell[state]['>'].nextstate = 0;

  state = 15;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].nextstate = 12;
    cell[state][chr].action = NULL;
  }
  cell[state]['s'].nextstate = 16;
  cell[state]['S'].nextstate = 16;
  cell[state]['>'].nextstate = 0;

  state = 16;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].nextstate = 12;
    cell[state][chr].action = NULL;
  }
  cell[state]['e'].nextstate = 17;
  cell[state]['E'].nextstate = 17;
  cell[state]['>'].nextstate = 0;

  state = 17;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].action = NULL;
    if( isspace(chr) )
      cell[state][chr].nextstate = 18;
    else
      cell[state][chr].nextstate = 12;
  }
  cell[state]['>'].nextstate = 0;

  state = 18;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].action = NULL;
    if( isspace(chr) )
      cell[state][chr].nextstate = 18;
    else
      cell[state][chr].nextstate = 12;
  }
  cell[state]['>'].nextstate = 0;
  cell[state]['h'].nextstate = 19;
  cell[state]['H'].nextstate = 19;

  state = 19;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].nextstate = 12;
    cell[state][chr].action = NULL;
  }
  cell[state]['r'].nextstate = 20;
  cell[state]['R'].nextstate = 20;
  cell[state]['>'].nextstate = 0;

  state = 20;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].nextstate = 12;
    cell[state][chr].action = NULL;
  }
  cell[state]['e'].nextstate = 21;
  cell[state]['E'].nextstate = 21;
  cell[state]['>'].nextstate = 0;

  state = 21;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].nextstate = 12;
    cell[state][chr].action = NULL;
  }
  cell[state]['f'].nextstate = 22;
  cell[state]['F'].nextstate = 22;
  cell[state]['>'].nextstate = 0;

  state = 22;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].action = NULL;
    if( isspace(chr) )
      cell[state][chr].nextstate = 22;
    else
      cell[state][chr].nextstate = 12;
  }
  cell[state]['>'].nextstate = 0;
  cell[state]['='].nextstate = 23;

  state = 23;
  for( chr = 0; chr < NCHARS; ++chr ) {
    cell[state][chr].action = NULL;
    if( isspace(chr) )
      cell[state][chr].nextstate = 23;
    else
      cell[state][chr].nextstate = 12;
  }
  cell[state]['>'].nextstate = 0;
  cell[state]['\"'].nextstate = 24;

  state = 24;
  for( chr = 0; chr < NCHARS; ++chr )
    if( isspace(chr) ) {
      cell[state][chr].nextstate = 12;
      cell[state][chr].action = reset;
    } else {
      cell[state][chr].nextstate = 24;
      cell[state][chr].action = store_url;
    }
  cell[state]['>'].nextstate = 0;
  cell[state]['>'].action = reset;
  cell[state]['\"'].nextstate = 12;
  cell[state]['\"'].action = have_base;

} /* InitHtmlParseTable */


/*=============================== Public Stuff ===============================*/

/*-----------------------------------------------------------------------*/
/* perform required initializations before html parsing can be performed */
/*-----------------------------------------------------------------------*/
void
InitHtmlParse( char *stopfile ) {

  InitHtmlParseTable();
  StopListInit( stopfile );

} /* InitHtmlParse */

/*-----------------------------------*/
/* initialize the html parser object */
/*-----------------------------------*/
HPP
CreateHtmlParser( void ) {

  HPP hpp;

  if( !(hpp = malloc(sizeof(HtmlParser))) ) {
    errlogf( ERN, "malloc failed in htmparse.c: InitHtmlParser\n" );
    exit( 1 );
  }

  hpp->crtstate = hpp->outchr = hpp->in_link = hpp->outcome = 0;
  hpp->stopw = StopReset();

  return( hpp );

} /* CreateHtmlParser */


/*----------------------------------------------------------------------------*/
/* parse html text from buffer, building a WordVec for the entire document,   */
/* and a LinkVec for the set of links on that document...                     */
/* Note: The current input character has to be converted to unsigned char in  */
/*       order to access entries in the table. Converting to int leads to     */
/*       negative indices, and therefore to seg faults when accessing entries.*/
/*----------------------------------------------------------------------------*/
void
ParseHtml( char *inbuf, int nchrs, HPP hpp, WordVec v, LinkVec lv ) {

  int i;

  for( i = 0; i < nchrs; ++i ) {

    /* execute action given current parser state and input character */
    if( cell[hpp->crtstate][(unsigned char)inbuf[i]].action )
      cell[hpp->crtstate][(unsigned char)inbuf[i]].action( hpp, inbuf[i] );

    /* examine and process the action result, if applicable */
    switch( hpp->outcome ) {
      case HAVE_URL:
        LinkVecAddURL( lv, hpp->outbuf );
        reset( hpp, '\0' );
        break;
      case HAVE_BASE:
        LinkVecChangeBase( lv, hpp->outbuf );
        reset( hpp, '\0' );
        break;
      case HAVE_WORD:
        if( !StopBoolIsStop(hpp->stopw) ) {
          if( WordVecInsert(v, hpp->outbuf) < 0 ) {
            errlogf( ERR, "ParseHtml(): insertion into wordvec failed\n" );
            exit( 1 );
          }
          if( hpp->in_link ) {	/* only for link words */
            LinkVecAddLinkWord( lv, hpp->outbuf );
          } else {	/* only for non-link words (nothing, for now...) */
	  ;
          }
        }
        reset( hpp, '\0' );
        break;
      default:
        break;
    }

    /* go to next state in parser */
    hpp->crtstate = cell[hpp->crtstate][(unsigned char)inbuf[i]].nextstate;

  } /* for */

} /* ParseHtml */


/*------------------------------------*/
/* delete html parser and return NULL */
/*------------------------------------*/
HPP
DeleteHtmlParser( HPP hpp ) {

  if( hpp )
    free( hpp );

  return( NULL );

} /* DeleteHtmlParser */
