/* This file is a component of the SurfAgent Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/********************************  linkvec.c  **********************************

    Purpose: data type: vector of html links.

    Notes:   This module implements a collection of html links (the url and
             the tfidf vector describing the words around the link)
             The number of words before / after the link that are considered
             to be significant for predicting the content of the document linked
             is called the window size and is learned using TD(0) with total
             discounting... (lambda =0 ).
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "wordvec.h"
#include "urlstr.h"
#include "linkvec.h"
#include "miscstr.h"
#include "errlog.h"



typedef struct _LinkListStruct {
  UrlContent url;		/* url pointed to */
  WordVec v;			/* tfidf vector describing the link */
  struct _LinkListStruct *next;
} LinkListStruct, *LinkList;


typedef struct _LinkVecStruct {
  UrlContent base_url;		/* base url of the document is given */
  LinkList links;		/* list of actual links */
} LinkVecStruct;


/*============================== Public functions ============================*/


/*----------------------------------------------------------------------------*/
/* Create a new link-vector object (the base url of the document is given)    */
/*----------------------------------------------------------------------------*/
LinkVec
LinkVecCreate( char *base ) {

  LinkVec lv;

  /* allocate new object */
  if( !(lv = malloc(sizeof(LinkVecStruct))) ) {
    errlogf( ERN, "LinkVecCreate(): malloc error\n" );
    exit( 1 );
  }

  /* copy base url */
  if( !(lv->base_url = parse_url(base)) ) {
    errlogf( ERR, "LinkVecCreate(): failed to create linkvec with base: %s\n",
                  base );
    exit( 1 );
  }

  /* init other members */
  lv->links = NULL;

  /* return new object */
  return( lv );

} /* LinkVecCreate */


/*----------------------------------------------------------------------------*/
/* Change the base url in the LinkVec (this is done when a <base ...>         */
/* statement was detected in the document                                     */
/*----------------------------------------------------------------------------*/
void
LinkVecChangeBase( LinkVec lv, char *newbase ) {

  UrlContent new_parsed_base;

  /* should we even bother */
  if( !lv )
    return;

  /* parse new base */
  if( !(new_parsed_base = parse_url(newbase)) ) {
    errlogf( ERR, "LinkVecChangeBase(): failed change to: %s\n", newbase );
    return;	/* leave old base in place for lack of something better */
  }

  /* ditch old base */
  DeleteUrlContent( lv->base_url );

  /* set new base */
  lv->base_url = new_parsed_base;

} /* LinkVecChangeBase */


/*----------------------------------------------------------------------------*/
/* Add a url to the linkvector - create a new node                            */
/* Note: url might be a relative path, and we need to do some handling...     */
/*----------------------------------------------------------------------------*/
void
LinkVecAddURL( LinkVec lv, char *url ) {

  LinkList newlink;

  /* should we even bother */
  if( !lv )
    return;

  /* create a new node and insert to front of links */
  if( !(newlink = malloc(sizeof(LinkListStruct))) ) {
    errlogf( ERN, "LinkVecAddURL(): malloc error\n" );
    exit( 1 );
  }
  newlink->next = lv->links;
  lv->links = newlink;

  if( is_qualified(url) )
    newlink->url = parse_url( url );
  else
    newlink->url = qualify_url( lv->base_url, url );

  if( !(newlink->v = WordVecNew()) ) {
    errlogf( ERR, "LinkVecAddURL(): failed to create wordvec\n" );
    exit( 1 );
  }

} /* LinkVecAddURL */


/*----------------------------------------------------------------------------*/
/* Add a link-word to the linkvector                                          */
/*----------------------------------------------------------------------------*/
void
LinkVecAddLinkWord( LinkVec lv, char *word ) {

  /* should we even bother */
  if( !lv )
    return;

  /* add the word to the last link */
  if( WordVecInsert(lv->links->v, word) < 0 ) {
    errlogf( ERR, "LinkVecAddLinkWord(): failed to insert into wordvector\n" );
    exit( 1 );
  }

} /* LinkVecAddLinkWord */


/*----------------------------------------------------------------------------*/
/* Extract the URLs of the links in lv into a StrList                         */
/* returns StrList on success, NULL on error                                  */
/*----------------------------------------------------------------------------*/
StrList
LinkVecGetURLs( const LinkVec lv ) {

  LinkList l;
  StrList retval;
  char *p;

  /* should we even bother */
  if( !lv ) {
    errlogf( ERR, "LinkVecGetURLs(): link vector is NULL!\n" );
    return( NULL);
  }

  if( !(retval = StrListCreate()) ) {
    errlogf( ERR, "LinkVecGetURLs(): failed to create StrList\n" );
    return( NULL );
  }

  for( l = lv->links; l; l = l->next ) {
    p = NULL;
    p = strsav( p, l->url->proto );
    p = strsav( p, "://" );
    p = strsav( p, l->url->hostport );
    p = strsav( p, l->url->path );
    if( StrListAppend(retval, p) != 0 ) {
      errlogf( ERR, "LinkVecGetURLs(): failed to insert URL into StrList\n" );
      StrListDestroy( retval );
      return( NULL );
    }
    freez( p );
  }

  return( retval );

} /* LinkVecGetURLs */


/*----------------------------------------------------------------------------*/
/* Delete a LinkVec object. return null (empty) object.                       */
/*----------------------------------------------------------------------------*/
LinkVec
LinkVecDelete( LinkVec lv ) {

  LinkList l, q;

  if( lv ) {

    /* delete list of links */
    for( l = lv->links; l; ) {
      q = l;
      l = l->next;
      DeleteUrlContent( q->url );
      WordVecDelete( q->v );
      free( q );
    }

    /* delete the rest */
    DeleteUrlContent( lv->base_url );

    free( lv );

  }

  return( NULL );

} /* LinkVecDelete */

