/* This file is a component of the SurfAgent Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
/********************************  profile.c  **********************************

    Purpose: Handling of user-profiling tasks.

    Notes:   This module maintains the user database, and offers methods for
             addition and retrieval of information to this database (with locks
             and all the works for use in a multithreaded environment)
**/

/*includes*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <dirent.h>
#include <pthread.h>

#include "strlist.h"
#include "wordvec.h"
#include "tfidf.h"
#include "linkvec.h"
#include "errlog.h"
#include "miscstr.h"

#include "htmparse.h"			/* for InitHtmlParse() */
#include "chat.h"			/* for get_and_analyze, SITE_MON */


/*defines (private - globals are in profile.h)*/

#define THR_INI 0.5			/* initial threshold for retrieval */
#define ALPHA   0.2			/* threshold learning rate */ 
#define EPSILON 0.07			/* exploration probability */
#define DOCW    0.25			/* document weight for vector sum */
#define TWAIT 14400		/* (4hrs)  wait interval for site monitor */

#define D_CNT "tfidf.d"			/* file names for the tfidf analyzer */
#define DF_VEC "tfidf.df"

#define STOPWORD_INI "stopword.ini"	/* file name for the stopword list */
#define CAT_THR_DIR "thresholds"	/* directory for category thresholds */

#define MAX_USERS 20			/* maximum number of users allowed */


/*macros (private)*/

/* uniformly distributed random double in the interval [0, 1) */
/* may be non-portable; only known to work on Linux and Solaris */
#define RZ1 ( (double)random() / (double)(((unsigned long)1 << 31) - 1) )


/*typedefs (private)*/

/* User Profile Category list element */
typedef struct _CategListStruct {
  char *c_name;
  WordVec c_vec;
  double c_thr;
  struct _CategListStruct *next;
} CategListStruct, *CategList;

/* Account List element */
typedef struct _UserListStruct {
  char *acct_name;
  pthread_rwlock_t acct_lock;		/* protects the ENTIRE account */
  CategList acct_categs;
  StrList acct_monlist;
  char *acct_monres;
  struct _UserListStruct *next;		/* once set, this never changes
					 * (insertion occurs at the front)
					 */
} UserListStruct, *UserList;


/*global variables (private, static)*/

static char *profiledir;	/* directory containing profile info... */
static char *tfidf_d, *tfidf_df;/* file names for tfidf analyzer */

static TfIdf tfidf;		/* tfidf analyzer */
static pthread_rwlock_t tfidf_lock;	/* protects analyzer */

static int nusers = 0;
static UserList userlist = NULL;/* list of users and their profiles */
static pthread_rwlock_t userlist_lock;	/* protects user list */


/*functions (private, static)*/

/*---------------------------------------------------------------------------*/
/* save category information to disk. We assume that the sub-directory given */
/* by CAT_THR_DIR exists; also, we assume we hold all the required locks to  */
/* mess in the user's dir. Returns 0 on success, -1 on error                 */
/*---------------------------------------------------------------------------*/
static int
save_categ( const char *user_id, const CategList categ ) {

  char wordfile[BUFSIZ], thrfile[BUFSIZ];

  sprintf( wordfile, "%s/%s/%s", profiledir, user_id, categ->c_name );
  sprintf( thrfile,  "%s/%s/%s/%s", profiledir, user_id,
                                    CAT_THR_DIR, categ->c_name );
  if( (WordVecSave(categ->c_vec, wordfile) < 0) ||
      (write_double_file(thrfile, categ->c_thr) < 0) ) {
    errlogf( ERR, "save_categ(): error saving category: %s:%s\n",
                  user_id, categ->c_name );
    return( -1 );
  }

  return( 0 );

} /* save_categ */


/*---------------------------------------------------------------------*/
/* create a new category node, with the name set to cat_name, an empty */
/*   word vector, threshold set to THR_INI, and next set to NULL.      */
/* returns pointer to node on success, NULL on failure                 */
/*---------------------------------------------------------------------*/
static CategList
new_categ( const char *cat_name ) {

  CategList new_c;

  /* create new node */
  if( !(new_c = malloc(sizeof(CategListStruct))) ) {
    errlogf( ERN, "new_categ(): malloc failed for new categ: %s\n", cat_name );
    return( NULL );
  }

  new_c->c_name = strdup( cat_name );
  if( !(new_c->c_vec = WordVecNew()) ) {
    errlogf( ERR, "new_categ(): error creating wordvector\n" );
    exit( 1 );
  }
  new_c->c_thr = THR_INI;
  new_c->next = NULL;

  return( new_c );

} /* new_categ */


/*----------------------------------------------------------------------------*/
/* create a new categ node and fill it with info from the named category file */
/* return ptr. to new category on success, NULL on failure                    */
/*----------------------------------------------------------------------------*/
static CategList
load_categ( const char *user_id, const char *cat_name ) {

  CategList new_c;
  char wordfile[BUFSIZ], thrfile[BUFSIZ];

  sprintf( wordfile, "%s/%s/%s", profiledir, user_id, cat_name );
  sprintf( thrfile,  "%s/%s/%s/%s", profiledir, user_id,
                                    CAT_THR_DIR, cat_name );

  /* create new node */
  if( !(new_c = malloc(sizeof(CategListStruct))) ) {
    errlogf( ERN, "load_categ(): malloc failed for new categ: %s:%s\n",
                  user_id, cat_name );
    return( NULL );
  }

  new_c->c_name = strdup( cat_name );
  new_c->c_vec = WordVecLoad( wordfile );
  new_c->c_thr = THR_INI;	/* just for good measure */
  read_double_file( thrfile, &(new_c->c_thr) );
  new_c->next = NULL;

  return( new_c );

} /* load_categ */


/*functions (public, to be found in profile.h)*/

/*----------------------------------------------------------------*/
/* Load in all existing user profile data (given the profiledir)  */
/* Parameters: char *profiledir_init : the location of the profile.   */
/* Returns: nothing.                                              */
/* Note: This code is executed before multithreaded action starts */
/*----------------------------------------------------------------*/
void
load_profile( char *profiledir_init ) {

  char pathbuf[BUFSIZ];
  DIR *dp = NULL;
  struct dirent *dfp;
  struct stat statbuf;
  UserList new_u;
  CategList new_c;

  /* set the local value of the profile location */
  profiledir = strdup( (profiledir_init) ? profiledir_init : "profile" );

  /* set the values for the tfidf files */
  sprintf( pathbuf, "%s/%s", profiledir, D_CNT );
  tfidf_d = strdup( pathbuf );
  sprintf( pathbuf, "%s/%s", profiledir, DF_VEC );
  tfidf_df = strdup( pathbuf );

  /* load tfidf analyzer and init its lock */
  pthread_rwlock_init( &tfidf_lock, NULL );
  if( !(tfidf = TfIdfLoad(tfidf_d, tfidf_df)) ) {
    errlogf( ERR, "load_profile(): failed to get tfidf object\n" );
    exit( 1 );
  }

  /* init userlist lock and prepare to load */
  pthread_rwlock_init( &userlist_lock, NULL );

  /* open profiledir */
  if( !(dp = opendir(profiledir)) ) {
    errlogf( ERN, "load_profile(): failed to open profile dir: %s\n",
                  profiledir );
    exit( 1 );
  }

  /* read all entries in the profile directory; subdirs are usernames */
  while( (dfp = readdir(dp)) ) {

    sprintf( pathbuf, "%s/%s", profiledir, dfp->d_name );/* full path to user */
    if( stat(pathbuf, &statbuf) != 0 ) {
      errlogf( ERN, "load_profile(): error getting info on %s (1)\n",
                    dfp->d_name );
      exit( 1 );
    }

    if( (statbuf.st_mode & S_IFDIR) &&		/* directories are users */
        (strcmp(dfp->d_name, ".") != 0) &&
        (strcmp(dfp->d_name, "..") != 0) ) {
      /* this one's a directory - it's a user */

      /* check availability */
      if( nusers >= MAX_USERS ) {
        errlogf( ERR, "load_profile(): too many accounts...\n" );
        exit( 1 );
      }
      ++nusers;

      /* create new user node */
      if( !(new_u = malloc(sizeof(UserListStruct))) ) {
        errlogf( ERN, "load_profile(): malloc for new user failed\n" );
        exit( 1 );
      }

      /* fill in new user node */
      new_u->acct_name = strdup( dfp->d_name );
      pthread_rwlock_init( &(new_u->acct_lock), NULL );
      new_u->acct_categs = NULL;
      new_u->acct_monlist = StrListCreate();
      new_u->acct_monres  = NULL;

      /* insert at front of userlist (dirnames are unique, no need to check) */
      new_u->next = userlist;
      userlist = new_u;

    } /* if( ... S_IFDIR ... ) */

  } /* while( (dfp = ...) ) */

  /* close profile dir */
  closedir( dp );

  /* read categories for each user */
  for( new_u = userlist; new_u; new_u = new_u->next ) {

    /* open user's directory */
    sprintf( pathbuf, "%s/%s", profiledir, new_u->acct_name );
    if( !(dp = opendir(pathbuf)) ) {
      errlogf( ERN, "load_profile(): failed to open user dir: %s\n",
                    new_u->acct_name );
      exit( 1 );
    }

    /* read all files in the user's directory */
    while( (dfp = readdir(dp)) ) {

      char user_pathbuf[BUFSIZ];

      sprintf( user_pathbuf, "%s/%s", pathbuf, dfp->d_name );
      if( stat(user_pathbuf, &statbuf) != 0 ) {
        errlogf( ERN, "load_profile(): error getting info on %s (2)\n",
                      dfp->d_name );
        exit( 1 );
      }
      if( statbuf.st_mode & S_IFREG ) {  /* file -> category or monitor list */

        if( strcmp(dfp->d_name, SITE_MON) ==0 ) {	/* monlist */
          if( StrListAppendFile(new_u->acct_monlist, user_pathbuf) != 0 ) {
            errlogf( ERR, "load_profile(): failed to init monitor list for "
                          " user: %s\n",
                          new_u->acct_name );
            exit( 1 );
          }
        } else {	/* create new category node */
        
          if( !(new_c = load_categ(new_u->acct_name, dfp->d_name)) ) {
            errlogf( ERR, "load_profile(): failed loading category: %s:%s\n",
                          new_u->acct_name, dfp->d_name );
            exit( 1 );
          }

          /* insert at front of catlist (names are unique, no need to check) */
          new_c->next = new_u->acct_categs;
          new_u->acct_categs = new_c;

        }
      }
    }

    /* close user's directory */
    closedir( dp );

  } /* end for( new_u ... ) */

  /* start parser and stop-word remover */
  sprintf( pathbuf, "%s/%s", profiledir, STOPWORD_INI );
  InitHtmlParse( pathbuf );

  errlogf( GPC, "Load Profile Done.\n\n" );

} /* load_profile */


/*--------------------------------------------------------------------------*/
/* find the user in the userlist - requires rdlock only. Also, we never     */
/* delete any users while the program is running (this can be done with     */
/* the program stopped, by simply deleting the corresponding directory)     */
/* and therefore once we've found a user and started working on their       */
/* account, there are no more conflicts to be avoided by using the locks... */
/* Note: we assume the user_id parameter is valid...                        */
/*--------------------------------------------------------------------------*/
static UserList
find_user( const char *user_id ) {

  UserList crt_user;

  errlogf( THR, "find_user(): request rdlock on user list\n" );
  pthread_rwlock_rdlock( &userlist_lock );
  errlogf( THR, "find_user(): now got rdlock on user list\n" );

  for( crt_user = userlist;
       crt_user && (strcmp(crt_user->acct_name, user_id) != 0);
       crt_user = crt_user->next );

  errlogf( THR, "find_user(): release rdlock on user list\n" );
  pthread_rwlock_unlock( &userlist_lock );

  return( crt_user );

} /* find_user */


/*-------------------------------------------------------------------------*/
/* create a new account - requires wrlock. The recommended usage of this   */
/* routine is for the situation when a find_user() failed and we want to   */
/* force the insertion of a new account. It is expected to be cheaper too, */
/* since we avoid having to use wrlock before we know the account is not   */
/* already there (which it is, most of the time)...                        */
/* Note: we assume the user_id parameter is valid...                       */
/*-------------------------------------------------------------------------*/
static UserList
add_user( char *user_id ) {

  UserList crt_user;
  char userdir[BUFSIZ], thrdir[BUFSIZ];

  /* prepare to write to the user list */
  errlogf( THR, "add_user(): request wrlock on user list\n" );
  pthread_rwlock_wrlock( &userlist_lock );
  errlogf( THR, "add_user(): now got wrlock on user list\n" );

  /* scan to make sure the account isn't already there */
  for( crt_user = userlist;
       crt_user && (strcmp(crt_user->acct_name, user_id) != 0);
       crt_user = crt_user->next );
  if( crt_user ) {	/* miraculously materialized, return it and be happy
			 * We just duplicated find_user() with a write lock,
			 * but it's a small price to pay at this low likelihood
			 */
    pthread_rwlock_unlock( &userlist_lock );
    return( crt_user );
  }

  /* make sure we have room for another user */
  if( nusers >= MAX_USERS ) {		/* no more users allowed */
    errlogf( ERR, "add_user(): Maximum number of users (%d) reached\n",
                  MAX_USERS );
    pthread_rwlock_unlock( &userlist_lock );
    return( NULL );
  }

  /* create new node */
  if( !(crt_user = malloc(sizeof(UserListStruct))) ) {
    errlogf( ERN, "add_user(): malloc failed\n" );
    exit( 1 );
  }

  /* increment user count */
  ++nusers;

  /* init account information */
  crt_user->acct_name = strdup( user_id );
  pthread_rwlock_init( &(crt_user->acct_lock), NULL );
  crt_user->acct_categs = NULL;
  crt_user->acct_monlist = StrListCreate();
  crt_user->acct_monres  = NULL;

  /* insert into userlist */
  crt_user->next = userlist;
  userlist = crt_user;

  /* create directory structure for user */
  sprintf( userdir, "%s/%s", profiledir, crt_user->acct_name );
  sprintf( thrdir,  "%s/%s", userdir, CAT_THR_DIR );
  if( (mkdir(userdir, S_IRWXU) != 0) || (mkdir(thrdir, S_IRWXU) != 0) ) {
    errlogf( ERN, "add_user(): error creating directory for new user: %s\n",
                  crt_user->acct_name );
    exit( 1 );
  }

  /* done writing to the userlist */
  errlogf( THR, "add_user(): release wrlock on user list\n" );
  pthread_rwlock_unlock( &userlist_lock );

  return( crt_user );

} /* add_user */


/*--------------------------------------------------------------------------*/
/* At the end of downloading a regular document, this function updates the  */
/*  tfidf object (inc d, add words to df). Also, given the user id, the     */
/*  list of user categories is returned...                                  */
/*--------------------------------------------------------------------------*/
char *
update_download( WordVec wv, char *user_id ) {

  UserList crt_user;
  CategList crt_categ;
  char buf[BUFSIZ], *categs = NULL;

  /* verify parameter sanity */
  if( !(wv && user_id) ) {
    errlogf( ERR, "update_download(): wv or user_id is NULL.\n" );
    exit( 1 );
  }

  /* convert wv from simple tf to tf-idf, updating tfidf object */
  errlogf( THR, "upadte_download(): request wrlock on tfidf\n" );
  pthread_rwlock_wrlock( &tfidf_lock );
  errlogf( THR, "update_download(): now got wrlock on tfidf\n" );
  if( TfIdfAnalyze(tfidf, wv, TFIDF_ANALYZE_UPDATE) < 0 ) {
    errlogf( ERR, "update_download(): tfidf analysis error.\n" );
    exit( 1 );
  }
  TfIdfSave( tfidf, tfidf_d, tfidf_df );	/* save changed tfidf to disk */
  errlogf( THR, "update_download(): release wrlock on tfidf\n" );
  pthread_rwlock_unlock( &tfidf_lock );

  /* find user */
  if( !(crt_user = find_user(user_id)) ) {
    errlogf( ERR, "update_download(): can't find user %s.\n", user_id );
    exit( 1 );
  }

  /* generate category list for this user (read lock) */
  errlogf( THR, "update_download(): request rdlock on user %s\n", user_id );
  pthread_rwlock_rdlock( &(crt_user->acct_lock) );
  errlogf( THR, "update_download(): now got rdlock on user %s\n", user_id );
  for( crt_categ = crt_user->acct_categs;
       crt_categ;
       crt_categ = crt_categ->next ) {
    /* the size of the numeric field is very important in order to get the
     * correct category name (see filters.c, function sa_update_categ() )
     * currently, the size is 6 ( "(X.XX)" ).
     */
    sprintf( buf, "\n<option> (%4.2f)%s",
                  WordVecSimilarity(crt_categ->c_vec, wv),
                  crt_categ->c_name );
    categs = strsav( categs, buf );
  } /* end for( crt_categ ... ) */
  errlogf( THR, "update_download(): release rdlock on user %s\n", user_id );
  pthread_rwlock_unlock( &(crt_user->acct_lock) );

  /* make sure categs is non-null */
  if( !categs )
    categs = strdup( "<sa_none>" );

  return( categs );

} /* update_download */


/*---------------------------------------------------------------------------*/
/* This function is called when the user gives feedback on a document. The   */
/* document wordvector (converted to tf-idf) is added to the user's profile. */
/*---------------------------------------------------------------------------*/
void
update_profile( WordVec wv, const char *user_id, const char *sa_categname ) {

  UserList crt_user;
  CategList crt_categ;
  WordVec oldvec;
  double similarity;

  /* verify parameter sanity */
  if( !(wv && user_id && sa_categname) ) {
    errlogf( ERR, "update_profile(): wv or user_id or sa_categname "
                  "is NULL.\n" );
    exit( 1 );
  }

  /* convert wv from simple tf to tf-idf, leaving tfidf object intact */
/* problem: document might have changed since last update, we'll encounter
   new words and tfidf_analyze will barf due to a div-by-zero error. For
   now, we have to use TFIDF_ANALYZE_UPDATE !!!
  pthread_rwlock_rdlock( &tfidf_lock );
  if( TfIdfAnalyze(tfidf, wv, TFIDF_ANALYZE_ONLY) < 0 ) {
    errlogf( ERR, "update_profile(): tfidf analysis failed\n" );
    exit( 1 );
  }
  pthread_rwlock_unlock( &tfidf_lock );
*/
  /* convert wv from simple tf to tf-idf, updating tfidf object */
  errlogf( THR, "upadte_profile(): request wrlock on tfidf\n" );
  pthread_rwlock_wrlock( &tfidf_lock );
  errlogf( THR, "upadte_profile(): now got wrlock on tfidf\n" );
  if( TfIdfAnalyze(tfidf, wv, TFIDF_ANALYZE_UPDATE) < 0 ) {
    errlogf( ERR, "update_profile(): tfidf analysis failed\n" );
    exit( 1 );
  }
  TfIdfSave( tfidf, tfidf_d, tfidf_df );	/* save changed tfidf to disk */
  errlogf( THR, "upadte_profile(): release wrlock on tfidf\n" );
  pthread_rwlock_unlock( &tfidf_lock );

  /* find user */
  if( !(crt_user = find_user(user_id)) ) {
    errlogf( ERR, "update_profile(): can't find user %s.\n", user_id );
    exit( 1 );
  }

  /* update profile category (write lock) */

  errlogf( THR, "update_profile(): request wrlock on user %s\n", user_id );
  pthread_rwlock_wrlock( &(crt_user->acct_lock) );
  errlogf( THR, "update_profile(): now got wrlock on user %s\n", user_id );

  /* find category in user account */
  for( crt_categ = crt_user->acct_categs;
       crt_categ && (strcmp(crt_categ->c_name, sa_categname) != 0);
       crt_categ = crt_categ->next );

  /* if category not found, insert */
  if( !crt_categ ) {
    if( !(crt_categ = new_categ(sa_categname)) ) {
      errlogf( ERR, "update_profile(): failed to create new category.\n" );
      exit( 1 );
    }
    crt_categ->next = crt_user->acct_categs;
    crt_user->acct_categs = crt_categ;
  }

  /* compute prior similarity for threshold learning and feedback */
  similarity = WordVecSimilarity( crt_categ->c_vec, wv );

  /* old categ. vector  +=  new doc. vector */
  oldvec = crt_categ->c_vec;
  crt_categ->c_vec = WordVecSumWtNrm( oldvec, 1.0, WV_NORMALIZE_ON,
                                          wv, DOCW, WV_NORMALIZE_ON );
  oldvec = WordVecDelete( oldvec );

  /* update cathegory threshold: move towards the prior similarity */
  crt_categ->c_thr += ALPHA * ( similarity - crt_categ->c_thr );

  save_categ( user_id, crt_categ );

  errlogf( THR, "update_profile(): release wrlock on user %s\n", user_id );
  pthread_rwlock_unlock( &(crt_user->acct_lock) );

} /* update_profile */


/*------------------------------------------------------------------------*/
/* This function is called upon obtaining the user account name from the  */
/* client. Its purpose is to check the validity of the login combination  */
/* Arguments: char *user_id - in - the account name to validate           */
/* Returns: (int)  1 if user is valid, 0 if not.                           */
/* Note: currently this function scans the list and creates a new account */
/* if not found.                                                          */
/*------------------------------------------------------------------------*/
int
valid_user( char *user_id ) {

  /* if the user_id has not been set, it's invalid... */
  if( !user_id )
    return( 0 );

  /* search userlist - cheap and likely to succeed, and uses rdlock only */
  if( find_user(user_id) )
    return( 1 );

  /* not found. Try to create new account before returning "invalid" */
  if( add_user(user_id) )
    return( 1 );

  /* failure: not in userlist, and failed to add */
  return( 0 );

} /* valid_user */


/*--------------------------------------*/
/* add a URL to the user's site monitor */
/* return 0 - OK, -1 - Error            */
/*--------------------------------------*/
int
add_monlist( char *user_id, char *url ) {

  UserList crt_user;
  char buf[BUFSIZ];
  int ret = 0;

  if( !(crt_user = find_user(user_id)) ) {
    errlogf( ERR, "add_monlist(): invalid user_id: %s\n", user_id );
    return( -1 );
  }

  /* we're changing stuff in the account - write lock */
  errlogf( THR, "add_monlist(): request wrlock on user %s\n", user_id );
  pthread_rwlock_wrlock( &(crt_user->acct_lock) );
  errlogf( THR, "add_monlist(): now got wrlock on user %s\n", user_id );

  /* if duplicate, just return success */
  if( StrListFind(crt_user->acct_monlist, url) == -1 ) {
    if( (ret = StrListAppend(crt_user->acct_monlist, url)) == 0 ) {
      sprintf( buf, "%s/%s/%s", profiledir, user_id, SITE_MON );
      ret = StrListSave( crt_user->acct_monlist, buf );
    }
  }

  errlogf( THR, "add_monlist(): release wrlock on user %s\n", user_id );
  pthread_rwlock_unlock( &(crt_user->acct_lock) );

  return( ret );

} /* add_monlist */


/*-------------------------------------------*/
/* delete a URL from the user's site monitor */
/* return 0 - OK, -1 - Error                 */
/*-------------------------------------------*/
int
del_monlist( char *user_id, char *url ) {

  UserList crt_user;
  char buf[BUFSIZ];
  int idx, ret = 0;

  if( !(crt_user = find_user(user_id)) ) {
    errlogf( ERR, "del_monlist(): invalid user_id: %s\n", user_id );
    return( -1 );
  }

  /*  we're changing stuff in the account - write lock */
  errlogf( THR, "del_monlist(): request wrlock on user %s\n", user_id );
  pthread_rwlock_wrlock( &(crt_user->acct_lock) );
  errlogf( THR, "del_monlist(): now got wrlock on user %s\n", user_id );

  if( (idx = StrListFind(crt_user->acct_monlist, url)) != -1 )
    if( (ret = StrListRemove(crt_user->acct_monlist, idx)) == 0 ) {
      sprintf( buf, "%s/%s/%s", profiledir, user_id, SITE_MON );
      ret = StrListSave( crt_user->acct_monlist, buf );
    }

  errlogf( THR, "del_monlist(): release wrlock on user %s\n", user_id );
  pthread_rwlock_unlock( &(crt_user->acct_lock) );

  return( ret );

} /* del_monlist */


/*-------------------------------------------------------------------------*/
/* create a string containing the site monitor url's in a html option list */
/*-------------------------------------------------------------------------*/
char *
get_monlist( char *user_id ) {

  UserList crt_user;
  int i;
  char *ret = NULL;

  if( !(crt_user = find_user(user_id)) ) {
    errlogf( ERR, "get_monlist(): invalid user_id: %s\n", user_id );
    return( NULL );
  }

  /* we're only reading stuff from the account - read lock */
  errlogf( THR, "get_monlist(): request rdlock on user %s\n", user_id );
  pthread_rwlock_rdlock( &(crt_user->acct_lock) );
  errlogf( THR, "get_monlist(): now got rdlock on user %s\n", user_id );

  for( i = 0; i < StrListSize(crt_user->acct_monlist); ++i ) {
    ret = strsav( ret, "\n<option> " );
    ret = strsav( ret, StrListPeek(crt_user->acct_monlist, i) );
  }

  errlogf( THR, "get_monlist(): release rdlock on user %s\n", user_id );
  pthread_rwlock_unlock( &(crt_user->acct_lock) );

  return( ret );

} /* get_monlist */


/*-------------------------------------------------------------------------*/
/* create a string containing the site monitor's results in a table        */
/*-------------------------------------------------------------------------*/
char *
get_monres( char *user_id ) {

  UserList crt_user;
  char *ret;

  if( !(crt_user = find_user(user_id)) ) {
    errlogf( ERR, "get_monres(): invalid user_id: %s\n", user_id );
    return( NULL );
  }

  /* we're only reading stuff from the account - read lock */
  errlogf( THR, "get_monres(): request rdlock on user %s\n", user_id );
  pthread_rwlock_rdlock( &(crt_user->acct_lock) );
  errlogf( THR, "get_monres(): now got rdlock on user %s\n", user_id );

  if( crt_user->acct_monres )
    ret = strdup( crt_user->acct_monres );
  else
    ret = NULL;

  errlogf( THR, "get_monres(): release rdlock on user %s\n", user_id );
  pthread_rwlock_unlock( &(crt_user->acct_lock) );

  return( ret );

} /* get_monres */


/*----------------------------------------------------------------------------*/
/* go through all users' monlists, and check the links on those sites against */
/* the respective user's profile. Insert the ones that pass a threshold into  */
/* the respective user's monres (result) list.                                */
/* Note: this function should be the main function of a distinct thread.      */
/*       multiple instances of such a thread can and should not be started... */
/*----------------------------------------------------------------------------*/
void *
site_monitor( void *arg ) {

  UserList crt_user;
  int i, j;
  CategList crt_categ;
  char buf[BUFSIZ], *new_monres;

  for( ; ; ) {

    /* sleep for the interval */
    sleep( TWAIT );

    errlogf( THR, "site_monitor(): request rdlock on user list\n" );
    pthread_rwlock_rdlock( &userlist_lock );
    errlogf( THR, "site_monitor(): now got rdlock on user list\n" );

    for( crt_user = userlist; crt_user; crt_user = crt_user->next ) {

      errlogf( THR, "site_monitor(): request rdlock on user %s\n",
                    crt_user->acct_name );
      pthread_rwlock_rdlock( &(crt_user->acct_lock) );
      errlogf( THR, "site_monitor(): now got rdlock on user %s\n",
                    crt_user->acct_name );

      new_monres = NULL;

      for( i = 0; i < StrListSize(crt_user->acct_monlist); ++i ) {

        LinkVec lv = NULL;
        StrList linkstr;

        /* get links from this monitored site */
        if( get_and_analyze(StrListPeek(crt_user->acct_monlist, i), &lv, NULL)
            != 0 ) {
          errlogf( ERR, "site_monitor(): could not read links from %s\n",
                        StrListPeek(crt_user->acct_monlist, i) );
          continue;
        }
        linkstr = LinkVecGetURLs( lv );
        LinkVecDelete( lv );
        if( !linkstr ) {
          errlogf( ERR, "site_monitor(): error converting lv to linkstr\n" );
          continue;
        }

        /* analyze each linked site */
        for( j = 0; j < StrListSize(linkstr); ++j ) {

          WordVec wv;
          char *p = NULL;

          /* get wordvector of linked site */
          if( get_and_analyze(StrListPeek(linkstr, j), NULL, &wv) != 0 ) {
            errlogf( ERR, "site_monitor(): could not read wv from %s\n",
                          StrListPeek(linkstr, j) );
            continue;
          }

          /* convert wv from simple tf to tf-idf, updating tfidf object */
          errlogf( THR, "site_monitor(): request wrlock on tfidf\n" );
          pthread_rwlock_wrlock( &tfidf_lock );
          errlogf( THR, "site_monitor(): now got wrlock on tfidf\n" );
          if( TfIdfAnalyze(tfidf, wv, TFIDF_ANALYZE_UPDATE) < 0 ) {
            errlogf( ERR, "site_monitor(): tfidf analysis failed\n" );
            exit( 1 );
          }
          TfIdfSave( tfidf, tfidf_d, tfidf_df );/* save changed tfidf to disk */
          errlogf( THR, "site_monitor(): release wrlock on tfidf\n" );
          pthread_rwlock_unlock( &tfidf_lock );

          /* compare wv to all user categories and decide whether to retain */
          for( crt_categ = crt_user->acct_categs;
               crt_categ;
               crt_categ = crt_categ->next ) {

            double similarity = WordVecSimilarity( crt_categ->c_vec, wv );

            /* if either the new doc. exceeds the threshold, or with an
             * exploration probability EPSILON*similarity, get the document...
             */
            if( (similarity > crt_categ->c_thr) ||
                (EPSILON * similarity > RZ1)
              ) {
              if( !p ) {	/* first insert the url */
                sprintf( buf, "<tr><td><a href=\"%s\">%s</a></td><td>",
                              StrListPeek(linkstr, j),
                              StrListPeek(linkstr, j) );
                p = strsav( NULL, buf );
              }
              sprintf( buf, "(%4.2f)%s ", similarity, crt_categ->c_name );
              p = strsav( p, buf );
            }

          }
          if( p ) {
            p = strsav( p, "</td></tr>\n" );
            if( !new_monres )
              new_monres = strsav( NULL, "<table border>\n" );
            new_monres = strsav( new_monres, p );
            free( p );
          }

          WordVecDelete( wv );

        } /* for( j ... ) */

        StrListDestroy( linkstr );

      } /* for( i ... ) */

      if( new_monres )
        new_monres = strsav( new_monres, "</table>\n" );

      if( new_monres || crt_user->acct_monres ) {
        errlogf( THR, "site_monitor(): release rdlock on user %s\n",
                      crt_user->acct_name );
        pthread_rwlock_unlock( &(crt_user->acct_lock) );
        errlogf( THR, "site_monitor(): request wrlock on user %s\n",
                      crt_user->acct_name );
        pthread_rwlock_wrlock( &(crt_user->acct_lock) );
        errlogf( THR, "site_monitor(): now got wrlock on user %s\n",
                      crt_user->acct_name );
        freez( crt_user->acct_monres );
        crt_user->acct_monres = new_monres;
      } /* else BOTH new_monres and crt_user->acct_monres are NULL
         * i.e., no further cleanup is necessary...
         */

      errlogf( THR, "site_monitor(): release **lock on user %s\n",
                    crt_user->acct_name );
      pthread_rwlock_unlock( &(crt_user->acct_lock) );

    } /* for( crt_user ... ) */

    errlogf( THR, "site_monitor(): release rdlock on user list\n" );
    pthread_rwlock_unlock( &userlist_lock );

  } /* for( ; ; ) */
  
  return( NULL );

} /* site_monitor */

