/* This file is a component of the SurfAgent Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/********************************  stopword.c  *********************************

    Purpose: Maintain a stoplist and offer services to check whether a given
             sequence of characters is on it...

    Notes:   This module is implemented using a trie. This will also work if
             uninitialized (i.e., the trie will be null and nothing will ever
             be identified as a stopword...
**/

/*includes*/

#include <stdio.h>
#include "stopword.h"
#include "errlog.h"


static TrieNode tn_stop = NULL;		/* will contain the stoplist */


/*====================================================================*/
/* will read the stopwords from stopfile into the global tn_stop trie */
/* must be called before multithreaded operation starts...            */
/*====================================================================*/
void
StopListInit( const char *stopfile ) {

  FILE *fp = NULL;	/* NULL assignment gets rid of a compiler warning */
  char buf[BUFSIZ];
  int err = 0;

  /* open stopfile */
  if( !stopfile ) {
    stopfile = "<no filename given>";
    err = 1;
  }
  if( err || !(fp = fopen(stopfile, "r")) ) {
    err = 1;
  }
  if( err ) {
    errlogf( ERN, "StopListInit(): error opening stopfile: %s; "
                  "No stoplist will be used...\n",
                  stopfile );
    return;
  }

  /* create the empty trie */
  if( !(tn_stop = MakeEmptyTrie()) ) {
    errlogf( ERN, "StopListInit(): error creating empty trie...\n" );
    exit( 1 );
  }

  /* load strings from file into the trie */
  while( fscanf(fp, "%s", buf) == 1 ) {
    buf[BUFSIZ-1] = '\0';	/* make sure string is properly terminated */
    if( TrieInsertStr(tn_stop, buf, TRIE_INS_SET, 1) < 0 ) {
      errlogf( ERR, "StopListInit(): error inserting string into trie\n" );
      exit( 1 );
    }
  }

  /* done with the stopfile */
  fclose( fp );

} /* stoplist_init */


/*=============================================================*/
/* reset a stopword test: return the root of the stopword trie */
/*=============================================================*/
TrieNode
StopReset( void ) {

  return( tn_stop );

} /* StopReset */


/*===========================================================================*/
/* advance one char in the stopword test. Return the next node (can be NULL) */
/*===========================================================================*/
TrieNode
StopNextChar( TrieNode tn, char c ) {

  return( TrieNextChar(tn, c) );

} /* StopNextChar */


/*============================================================================*/
/* See if state tn is terminal, i.e. if the word scanned so far is a stopword */
/*============================================================================*/
int
StopBoolIsStop( TrieNode tn ) {

  return( TrieStringCount(tn) );

} /* StopBoolIsStop */
