/* This file is a component of the SurfAgent Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*******************************   strlist.c   *********************************

    Purpose: String list abstract data type implementation.

    Notes:   This module implements a straightforward string ordered list
             abstract data type.  It is optimized for appending and deleting
             from the end of the list.  Since they are ordered lists, string
             lists may be sorted, and their members are addressed by ordinal
             position (starting from 0).
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "errlog.h"

#include "strlist.h"

/******************************************************************************/
/******************   Private Defines and Data Structures   *******************/

#define EOS                          '\0'

#define INCREMENT                     32    /* increase size by this much */
#define MAX_LINE                     128    /* when reading text files */

typedef struct _StrListStruct {
  short size;             /* current length of the list */
  short max_size;         /* room for this many strings */
  char **string;          /* the string array */
} StrListStruct;


/**********************   Private Routine Declarations   *********************/


/*FN**************************************************************************

         ExpandArray( list )

   Returns: int -- 0 on success, -1 error

   Purpose: Increase the string array to hold more data

   Plan:    Part 1: Increase the maximum list size to its new value
            Part 2: Allocate a new chunk of memory
            Part 3: Return an indication of success

   Notes:   None

   Param:   StrList list - in: string list whose string array is enlarged;

**/

static int
ExpandArray( StrList list ) {

  /* Part 1: Increase the maximum list size to its new value */
  list->max_size += INCREMENT;

  /* Part 2: Allocate a new chunk of memory */
  list->string = realloc( list->string, list->max_size * sizeof(char *) );
  if( !(list->string) ) {
    errlogf( ERN, "ExpandArray(): realloc error\n" );
    exit( 1 );
  }

  /* Part 3: Return an indication of success */
  return( (list->string) ? 0 : -1 );

} /* ExpandArray */


/*FN**************************************************************************

         ISort( string, lb, ub )

   Returns: void

   Purpose: Insertion sort a string array forward using strcmp ordering

   Plan:    Part 1: Put smallest in place as a sentinal
            Part 2: Insert as necessary

   Notes:   None

   Param.:  char **string - in/out: string array sorted;
            int lb, ub    - in: array bounds for sort;

**/

static void
ISort( char **string, int lb, int ub ) {

  int i,j;           /* for scanning through the list */
  char *tmp;         /* for swaps */

  /* Part 1: Put smallest in place as a sentinal */
  for( j = lb, i = lb + 1; i <= ub; i++ )
    if( strcmp(string[j],string[i]) > 0 )
      j = i;
  tmp = string[lb];
  string[lb] = string[j];
  string[j] = tmp;

  /* Part 2: Insert as necessary */
  for( i = lb + 2; i <= ub; i++ ) {
    tmp = string[i];
    for( j = i; strcmp(string[j-1], tmp) > 0; j-- )
      string[j] = string[j-1];
    string[j] = tmp;
  }

} /* ISort*/


/*FN**************************************************************************

         QSort( string, lb, ub )

   Returns: void

   Purpose: Quicksort an array of strings forward using strcmp ordering

   Plan:    Part 1: Use insertion sort of the list is short
            Part 2: Do median of three pivot value selection
            Part 3: Put the pivot out of the way at the top
            Part 5: Swap the pivot back into the mid of the list
            Part 6: Recursively sort the sublists

   Notes:   Standard quicksort function with the two main enhancements:
            median of three partitioning to find a good pivot value,
            and sorting small arrays with insertion sort.

   Param.:  char **string - in/out: string array sorted;
            int lb, ub     - in: array bounds for sort;

**/

static void
QSort( char **string, int lb, int ub ) {

  int lft;           /* list pointer that closes from the left */
  int rgt;           /* list pointer that closes from the right */
  int mid;           /* index of the median of three value */
  char *tmp;         /* for string pointer swaps */
  char *pivot;       /* the pivot value string */

  /* Part 1: Use insertion sort of the list is short */
  if( ub-lb < 12 ) {
    ISort( string, lb, ub );
    return;
  }

  /* Part 2: Do median of three pivot value selection */
  mid = ( lb + ub ) / 2;
  if( strcmp(string[mid], string[lb]) < 0 ) {
    tmp = string[mid];
    string[mid] = string[lb];
    string[lb] = tmp;
  }
  if( strcmp(string[ub], string[mid]) < 0 ) {
    tmp = string[mid];
    string[mid] = string[ub];
    string[ub] = tmp;
  }
  if( strcmp(string[mid], string[lb]) < 0 ) {
    tmp = string[mid];
    string[mid] = string[lb];
    string[lb] = tmp;
  }

  /* Part 3: Put the pivot out of the way at the top */
  tmp = string[mid];
  string[mid] = string[ub-1];
  string[ub-1] = tmp;

  /* Part 4: Partition around the pivot value */
  lft = lb;
  rgt = ub-1;
  pivot = string[ub-1];
  do {

    do {
      lft++;
    } while( strcmp(string[lft], pivot) < 0 );

    do {
      rgt--;
    } while ( strcmp(pivot, string[rgt]) < 0 );

    tmp = string[lft];
    string[lft] = string[rgt];
    string[rgt] = tmp;

  } while( lft < rgt );

  /* Part 5: Swap the pivot back into the mid of the list */
  string[rgt] = string[lft];
  string[lft] = string[ub-1];
  string[ub-1] = tmp;

  /* Part 6: Recursively sort the sublists */
  QSort( string, lb, lft-1 );
  QSort( string, rgt+1, ub );

} /* QSort */


/**********************   Public Routine Definitions   ***********************/

/*FN***************************************************************************

         StrListAppend( list, string )

   Returns: int - 0 on success, -1 on failure

   Purpose: Place a string on the end of a string list

   Plan:    Part 1: Standard parameter sanity check
            Part 2: Expand the list as necessary
            Part 3: Append the new string to the tail

   Notes:   None

   Param.:  StrList list - in/out: list appended to;
            char *string - in: the appended string;

**/

int
StrListAppend( StrList list, const char *string ) {

  /* Part 1: Standard parameter sanity check */
  if( !list || !string )
    return( -1 );

  /* Part 2: Expand the list as necessary */
  if( (list->size == list->max_size) && (ExpandArray(list) != 0) )
    return( -1 );

  /* Part 3: Append the new string to the tail */
  list->string[list->size] = strdup( string );
  ++(list->size);

  return( 0 );

} /* StrListAppend */


/*FN***************************************************************************

         StrListRemove( list, idx )

   Returns: int - 0 on success, -1 on failure

   Purpose: Throw out an arbitrary string from a string list

   Plan:    Part 1: Standard parameter sanity check
            Part 2: De-allocate string
            Part 3: Slide remaining strings in place
            Part 4: Adjust size

   Notes:   if the order didn't matter (which I'm not sure yet), we could just
            copy the last one over the deleted one and get done with it without
            sliding everything. Remains to be seen...

   Param.:  StrList list - in/out: list to ditch from;
            int idx - in: the index of the string to be removed

**/

int
StrListRemove( StrList list, int idx ) {

  int i;

  /* Part 1: Standard parameter sanity check */
  if( !list || (list->size < 1) || (idx < 0) || (idx >= list->size) )
    return( -1 );

  /* Part 2: De-allocate string */
  free( list->string[idx] );

  /* Part 3: Slide remaining strings in place */
  for( i = idx + 1; i < list->size; ++i )
    list->string[i-1] = list->string[i];
  list->string[i] = NULL;

  /* Part 4: Adjust size */
  --(list->size);

  return( 0 );

} /* StrListRemove */


/*FN***************************************************************************

         StrListDequeue( list )

   Returns: int - 0 on success, -1 on failure

   Purpose: Throw out a string from the start of a string list

   Plan:    Call StrListRemove with idx = 0;

   Notes:   None

   Param.:  StrList list - in/out: list to ditch from;

**/

int
StrListDequeue( StrList list ) {

  return( StrListRemove(list, 0) );

} /* StrListDequeue */


/*FN***************************************************************************

         StrListPop( list )

   Returns: int - 0 on success, -1 on failure

   Purpose: Throw out a string from the end of a string list

   Plan:    Part 1: Standard parameter sanity check
            Part 2: Adjust size
            Part 3: De-allocate last string

   Notes:   None.

   Param.:  StrList list - in/out: list to remove from;

**/

int
StrListPop( StrList list ) {

  return( StrListRemove(list, list->size - 1) );

} /* StrListPop */


/*FN***************************************************************************

         StrListAppendFile( list, filename )

   Returns: int - 0 on success, -1 on failure

   Purpose: Place all lines from a file on the end of a string list

   Plan:    Part 1: Standard parameter sanity check
            Part 2: Expand the list as necessary
            Part 3: Append the new string to the tail

   Notes:   Uses StrListAppend() to reduce redundancy (GLS)

   Param.:  StrList list   - in/out: list appended to;
            char *filename - in: the appended file;

**/

int
StrListAppendFile( StrList list, char *filename ) {

  FILE *file;            /* file handle for the text input file */
  char buffer[MAX_LINE]; /* for storing text input file lines */

  /* Part 1: Standard parameter sanity check */
  if( !list || !filename ) {
    errlogf( ERR, "StrListAppendFile(): invalid parameters!\n" );
    return( -1 );
  }

  /* Part 2: Open the text input file; check for error */
  if( !(file = fopen(filename,"r")) ) {
    errlogf( ERN, "StrListAppendFile(): could not open file: %s\n", filename );
    return( -1 );
  }

  /* Part 3: Append to the list, checking for errors */
  while( fscanf(file, "%s", buffer) == 1 ) {

    buffer[MAX_LINE-1] = EOS;

    if( StrListAppend(list, buffer) != 0 ) {
      fclose( file );
      errlogf( ERR, "StrListAppendFile(): failed to add string: %s\n", buffer );
      return( -1 );
    }
  }

  /* Part 4: Close the text input file */
  fclose( file );

  return( 0 );

} /* StrListAppendFile */


/*FN***************************************************************************

         StrListCreate()

   Returns: StrList -- a new structure, or NULL on failure

   Purpose: Allocate and initialize a new string list structure

   Plan:    Part 1: Allocate space for the string list object
            Part 2: Initialize the structure fields
            Part 3: Return the new string list

   Notes:   None

   Param.:  None

**/

StrList
StrListCreate() {

  StrList list;  /* the new list returned */

  /* Part 1: Allocate space for the string list object */
  if ( !(list = malloc(sizeof(StrListStruct))) ) {
    errlogf( ERN, "StrListCreate(): malloc error\n" );
    exit( 1 );
  }

  /* Part 2: Initialize the structure fields */
  list->string = NULL;
  list->size = list->max_size = 0;
  if ( ExpandArray(list) != 0 ) {
    free( list );
    return( NULL );
  }

  /* Part 3: Return the new string list */
  return( list );

} /* StrListCreate */


/*FN***************************************************************************

         StrListDup( list )

   Returns: StrList -- a new structure, equal to the original, or NULL on
                       failure

   Purpose: Allocate and initialize a new string list structure

   Plan:    Part 1: Allocate space for the string list object
            Part 2: Copy the structure fields
            Part 3: Return the new string list

   Notes:   None

   Param.:  StrList list - in: original string list to duplicate

**/

StrList
StrListDup( StrList list ) {

  StrList newlist;  /* the new list returned */
  int i;

  /* Part 1: Allocate space for the string list object */
  newlist = malloc( sizeof(StrListStruct) );
  newlist->string = malloc( list->size * sizeof(char *) );
  if( !newlist || !(newlist->string) ) {
    errlogf( ERN, "StrListDup(): malloc error\n" );
    exit( 1 );
  }

  /* Part 2: Copy the structure fields */
  newlist->size = newlist->max_size = list->size;
  for( i = 0; i < newlist->size; ++i )
    newlist->string[i] = strdup( list->string[i] );

  /* Part 3: Return the new string list */
  return( newlist );

} /* StrListDup */


/*FN***************************************************************************

         StrListDestroy( list )

   Returns: void

   Purpose: Deallocate the space used for a string list

   Plan:    Part 1: Standard parameter sanity check
            Part 2: Free all the space

   Notes:   None

   Param.:  StrList list - in: the list destroyed;

**/

void
StrListDestroy( StrList list ) {

  int i;    /* for scanning through the list */

  /* Part 1: Standard parameter sanity check */
  if( !list )
    return;

  /* Part 2: Free all the space */
  for( i = 0; i < list->size; i++ )
    free( (list->string[i]) );

  free( list->string );

  free( list );

} /* StrListDestroy */


/*FN***************************************************************************

         StrListEqual( list1, list2 )

   Returns: int -- 0 if the lists are equivalent, -1 otherwise

   Purpose: See if two lists have identical elements

   Plan:    Part 1: Say not equal if the parameters are bad
            Part 2: Say not equal if sizes are different
            Part 3: Compare lists element by element
            Part 4: Say equal if everything checks out

   Notes:   None

   Param.:  StrList list1, list2 - in: lists compared;

**/

int
StrListEqual( StrList list1, StrList list2 ) {

  int i;  /* for scanning through the lists */

  /* Part 1: Say not equal if the parameters are bad */
  if( !list1 || !list2 )
    return( -1 );

  /* Part 2: Say not equal if sizes are different */
  if( list1->size != list2->size )
    return( -1 );

  /* Part 3: Compare the lists element by element */
  for( i = 0; i < list1->size; i++ )
    if( strcmp(list1->string[i], list2->string[i]) != 0 )
      return( -1 );

  /* Part 5: Say equal if everything checks out */
  return( 0 );

} /* StrListEqual */


/*FN***************************************************************************

         StrListPeek( list, index )

   Returns: char * -- pointer to the requested string; NULL on error

   Purpose: Peek a string by its list index

   Plan:    Part 1: Standard parameter sanity check
            Part 2: Return the requested string

   Notes:   Note that this function is a hole in the data type encapsulation:
            it should return a copy, but this would force the consumer to
            deallocate the string.  Design call.

   Param.:  StrList list - in: list retrieved from;
            int index    - in: which string to fetch;

**/

char *
StrListPeek( StrList list, int index ) {

  /* Part 1: Standard parameter sanity check */
  if( !list || (index < 0) || (list->size <= index) )
    return( NULL );

  /* Part 2: Return the requested string */
  return( list->string[index] );

} /* StrListPeek */


/*FN***************************************************************************

         StrListSize( list )

   Returns: int -- the size of the list, -1 on error

   Purpose: Grab the list size

   Plan:    Return the list size field

   Notes:   None

   Param.:  StrList list - in: list queried;

**/

int
StrListSize( StrList list ) {

  if( !list )
    return( -1 );
  else
    return( list->size );

} /* StrListSize */


/*FN***************************************************************************

         StrListSort( list )

   Returns: void

   Purpose: Sort a single string list using strcmp ordering

   Plan:    Part 1: Do parameter sanity checks, then sort

   Notes:   None

   Param.:  StrList list - in: list sorted;

**/

void
StrListSort( StrList list ) {

  /* Part 1: Do parameter sanity checks, then sort */
  if( !list )
    return;

  QSort( list->string, 0, list->size-1 );

} /* StrListSort */


/*FN***************************************************************************

         StrListUnique( list )

   Returns: void

   Purpose: Sort a single string list using strcmp ordering, then remove
            duplicates.

   Plan:    Part 1: Do parameters sanity checks
            Part 2: Sort the list
            Part 3: Remove duplicate strings

   Notes:   None

   Param.:  StrList list - in/out: list sortedand uniqued;

**/

void
StrListUnique( StrList list ) {

  int i, j;   /* counters for copying down over duplicates */

  /* Part 1: Do parameter sanity checks */
  if( !list )
    return;

  /* Part 2: Sort the list */
  QSort( list->string, 0, list->size-1 );

  /* Part 3: Remove duplicate strings */
  if( list->size > 1 ) {
    for( j = 0, i = 1; i < list->size; i++ ) {
      if( strcmp(list->string[i], list->string[j]) == 0 )
        free( list->string[j] );
      else
        ++j;
      if( j < i )
        list->string[j] = list->string[i];
    }
    list->size = j + 1;
  }

} /* StrListUnique */


/*FN***************************************************************************

         StrListFind( list, string )

   Returns: int - index of the first element matching string

   Purpose: Find whether string is contained in list, and return its index.

   Plan:    Part 1: Do parameters sanity checks
            Part 2: Search the list

   Notes:   None

   Param.:  StrList list - in/out: list sortedand uniqued;

**/

int
StrListFind( StrList list, char *string ) {

  int i;

  /* Part 1: Do parameter sanity checks */
  if( !list )
    return( -1 );

  /* Part 2: Search the list */
  for( i = 0; i < list->size; ++i )
    if( strcmp( list->string[i], string ) == 0 )
      return( i );

  /* not found */
  return( -1 );

} /* StrListFind */


/*FN***************************************************************************

         StrListSave( list, filename )

   Returns: 0 on success, -1 on failure

   Purpose: Save all strings in the list as lines in a file.
   Notes:   None
   Param.:  StrList list - in: list sortedand uniqued;
            char *string - in: the name of the file to save to;

**/

int
StrListSave( const StrList list, const char *filename ) {

  FILE *fp;
  int i;

  /* Do parameter sanity checks */
  if( !list || list->size < 0 ) {
    errlogf( ERR, "StrListSave(): invalid string list argument\n" );
    return( -1 );
  }

  /* Open the file (overwrite) */
  if( !(fp=fopen(filename, "w")) ) {
    errlogf( ERN, "StrListSave(): could not open file: %s\n", filename );
    return( -1 );
  }

  /* Write to file */
  for( i = 0; i < list->size; ++i )
    fprintf( fp, "%s\n", list->string[i] );

  /* Done. Close and report success */
  fclose( fp );
  return( 0 );

} /* StrListSave */

