/* This file is a component of the SurfAgent Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/********************************   tfidf.c   *********************************

    Purpose: Word vector abstract data type implementation.

    Notes:   This module implements a word-vector with word counts. It is
             intended for frequent insertions and updates, and therefore
             the underlying structure is a list sorted alphabetically.
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "errlog.h"

#include "miscstr.h"	/* for read_double_file(), write_double_file() */

#include "wordvec.h"
#include "tfidf.h"
#include "hash.h"


  /********************* TF-IDF analyzer structure *********************/
  /* An element of this type contains the data needed by TF-IDF        */
  /* analysis of a document represented by a word vector of counts.    */
  /* The count vector is transformed into a TF-IDF vector, and the     */
  /* internals of the analyzer are updated such that future document   */
  /* analysis is dependent on the hystory of documents seen.           */
typedef struct _TfIdfStruct {
  double d;			/* total number of documents analyzed so far */
  HashTable df;			/* document frequency trie */
} TfIdfStruct;


/*******************************************************************************
    TfIdfInit( )

    Purpose:   Create a new TF-IDF analyzer object.
    Returns:   TfIdf     - pointer to the newly created and initialized object
    Arguments: none
**/
TfIdf
TfIdfInit( void ) {

  TfIdf tfidf;

  /* create new object */
  if( !(tfidf=malloc(sizeof(TfIdfStruct))) ) {
    errlogf( ERN, "TfIdfInit(): malloc error\n" );
    exit( 1 );
  }

  /* initialize the object */
  tfidf->d  = 0.0;
  if( !(tfidf->df = CreateHashTable()) ) {
    errlogf( ERR, "TfIdfInit(): error creating hash table\n" );
    exit( 1 );
  }

  /* return the object */
  return( tfidf );

} /* TfIdfInit */


/*******************************************************************************
    TfIdfLoad( )

    Purpose:   Load a TF-IDF analyzer object (from two files: D and DF).
    Returns:   TfIdf - pointer to the newly created and initialized object
    Arguments: char *dname - in: the name of the file containing D;
               char *dfname - in: the name of the file containing DF;
**/
TfIdf
TfIdfLoad( char *dname, char *dfname ) {

  TfIdf tfidf;
  FILE *fp;
  double dcount;

  /* create new object */
  if( !(tfidf=TfIdfInit()) ) {
    errlogf( ERR, "TfIdfLoad(): error creating object, ret. NULL\n" );
    return( NULL );
  }

  /* open D file: on failure, resort to the clean new object created above */
  if( !(fp=fopen(dname, "r")) ) {
    errlogf( ERN, "TfIdfLoad(): error opening D count file %s\n", dname );
    return( tfidf );
  }

  /* read the file: it's one single double number */
  if( fscanf(fp, "%lf", &dcount) != 1 ) {
    errlogf( ERN, "TfIdfLoad(): error reading D count from file %s\n", dname );
    fclose( fp );
    return( tfidf );
  }
  tfidf->d = dcount;
  fclose( fp );

  /* now get DF */
  tfidf->df = HashLoad(dfname, tfidf->df);

  return( tfidf );

} /* TfIdfLoad */


/*******************************************************************************
    TfIdfSave( )

    Purpose:   Save a TF-IDF analyzer object (to two files: D and DF).
    Returns:   int - 0 on success, -1 on error (FIXED !!!)
    Arguments: TfIdf tfidf - in: the object to be saved;
               char *dname - in: the name of the file to contain D;
               char *dfname - in: the name of the file to contain DF;
    Notes: Saving doubles with fprintf loses precision, but is portable. If this
           were a single-platform implementation, it would make more sense to
           use write and sizeof(double)... (also see the load function)
**/
int
TfIdfSave( TfIdf tfidf, char *dname, char *dfname ) {

  if( write_double_file(dname, tfidf->d) < 0 ) {
    errlogf( ERR, "TfIdfSave(): error writing D to file: %s\n", dname );
    return( -1 );
  }

  /* now save DF to its file, and return the result */
  return( HashSave(tfidf->df, dfname) );

} /* TfIdfSave */


/*******************************************************************************
    TfIdfDelete( tfidf )

    Purpose:   Delete and deallocate a TF-IDF analyzer object.
    Returns:   TfIdf     - NULL pointer, so that we can use this function like
                           this:  tfidf = TfIdfDelete( tfidf );  - and tfidf
                           would end up being NULL after having its contents
                           deallocated;
    Arguments: TfIdf tfidf - in: the object to be deleted;
**/
TfIdf
TfIdfDelete( TfIdf tfidf ) {

  if( tfidf ) {

    /* deallocate contents */
    DeleteHashTable( tfidf->df );		/* deallocate the word vector */

    /* deallocate object */
    free( tfidf );

  }

  /* return NULL (the new value for tfidf) */
  return( NULL );

} /* TfIdfDelete */


/*******************************************************************************
    TfIdfAnalyze( tfidf, tf, update )

    Purpose:   Perform TF-IDF analysis on the word vector tf using tfidf object.
    Returns:   int         - 0 on success, -1 on error
    Arguments: TfIdf tfidf - in: the object to be used/updated in the analysis;
               WordVec tf  - in/out: word vector containing the term count of
                             the current document to be analyzed
               int update  - in: update the tfidf analyzer if this is true...
    Comments:  The vector v is not modified pointer-wise, only the count
               entries are changed to reflect TF-IDF.
**/
int
TfIdfAnalyze( TfIdf tfidf, WordVec tf, int update ) {

  WordList p;
  double sum = 0.0;
  int df_count;

  /* Verify parameter sanity */
  if( !(tfidf && tf) ) {
    errlogf( ERR, "TfIdfAnalyze(): NULL parameter (tfidf=%p, tf=%p)\n",
                  tfidf, tf );
    return( -1 );
  }

  /* increment total number of documents scanned so far with tfidf */
  if( update )
    tfidf->d += 1.0;

  /* go through all words in tf; update df(word) and change count to TF-IDF */
  for( p = tf->wordlist; p; p = p->next ) {

    /* update df(word). If no update, inc. w. 0.0 and read crt. value */
    df_count = HashInsStr( tfidf->df, p->word, HASH_INS_INC, 1 );

    /* Sanity check !!!! */
    if( df_count < 1 ) {
      errlogf( ERR, "TfIdfAnalyze(): division by 0 error...\n"
                    "  df_count came back to be 0 for word %s.\n"
                    " Update was %d\n\n",
                    p->word, update );
      return( -1 );
    }

    /* change p->count from absolute to TF-IDF */
    p->count *= log( tfidf->d / (double)df_count );

    sum += p->count * p->count;
  }

  tf->length = sqrt( sum );

  /* done, return success */
  return( 0 );

} /* TfIdfAnalyze */
