/* This file is a component of the SurfAgent Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/**********************************  trie.c  ***********************************

    Purpose: Offer services related to a trie data structure (insertion and
             lookup of strings, creation of new empty trie, deletion of a
             trie.

    Notes:   none.
**/

/*includes*/

#include <stdio.h>
#include "trie.h"
#include "errlog.h"
#include "miscstr.h"


/* trie of case-sensitive strings */
typedef struct _TrieNode {
  int count;				/* string occurrence counter */
  struct _TrieNode *next[NCHARS];	/* where each possible char leads us */
} TrieNodeStruct;


/*===================*/
/* Create empty trie */
/*===================*/
TrieNode
MakeEmptyTrie( void ) {

  TrieNode tn;

  if( !(tn = zalloc(sizeof(TrieNodeStruct))) )
    errlogf( ERN, "MakeEmptyTrie(): failed to allocate trie node\n" );

  return( tn );
} /* MakeEmptyTrie */


/*=======================================================================*/
/* Recursively insert a string into a trie;                              */
/*  flg controls whether the count is set to the parameter cnt or simply */
/*      incremented by 1 (in which case cnt is ignored)                  */
/*  the return value is the new count (>0) on success, and -1 on failure */
/*=======================================================================*/
int
TrieInsertStr( TrieNode tn, const char *str, int flg, int cnt ) {

  /* parameter sanity check */
  if( !tn || !str ) {
    errlogf( ERR, "TrieInsertStr(): invalid parameter (tn=%p, str=%p)\n",
                  tn, str );
    return( -1 );
  }

  if( str[0] == '\0' )		/* a string ends here - this is a final state */
    switch( flg ) {
      case TRIE_INS_INC:
        ++(tn->count);		/* cnt parameter is ignored in this case */
        return( tn->count );
      case TRIE_INS_SET:
        tn->count = cnt;
        return( tn->count );
      default:
        errlogf( ERR, "TrieInsertStr(): bad flg value: %d\n", flg );
        return( -1 );
    }

  /* make sure the next node corresponding to str[0] exists */
  if( !(tn->next[(unsigned char)str[0]]) &&
      !(tn->next[(unsigned char)str[0]] = zalloc(sizeof(TrieNodeStruct))) ) {
    errlogf( ERN, "TrieInsertStr(): failed to allocate new node\n" );
    exit( 1 );
  }

  /* insert rest of string into corresponding next node */
  return( TrieInsertStr(tn->next[(unsigned char)str[0]], str + 1, flg, cnt) );

} /* TrieInsertStr */


/*================================================*/
/* Load a trie from a file.                       */
/*================================================*/
TrieNode
TrieLoad( char *filename, TrieNode tn ) {

  FILE *fp;
  char buf[BUFSIZ];
  int cnt;

  /* create empty trie */
  if( !tn && !(tn = MakeEmptyTrie()) ) {
    errlogf( ERR, "TrieLoad(): error creating new trie.\n" );
    exit( 1 );
  }

  /* open source file */
  if( !(fp=fopen(filename, "r")) ) {
    errlogf( ERN, "TrieLoad(): error opening word vector file %s\n",
                  filename );
    return( tn );
  }

  /* read the file: each line is a pair of count and word */
  while( fscanf(fp, "%d%s", &cnt, buf) == 2 ) {
    buf[BUFSIZ-1] = '\0';	/* make sure string is terminated */
    if( TrieInsertStr(tn, buf, TRIE_INS_SET, cnt) < 0 ) {
      errlogf( ERR, "TrieLoad(): Error adding string %s from file %s\n",
                    buf, filename );
      exit( 1 );
    }
  }

  /* close the input file when done */
  fclose( fp );

  /* return the resulting trie */
  return( tn );

} /* TrieLoad */


/*===============================================================*/
/* preorder traversal of a trie, at depth d, writing stuff to fp */
/*===============================================================*/
static void
TriePreorder( TrieNode tn, char *buf, int d, FILE *fp ) {

  int i;

  /* parameter sanity */
  if( d >= BUFSIZ ) {
    errlogf( ERR, "TriePreorder(): buffer overflow !\n" );
    exit( 1 );
  }
  if( !tn )
    return;

  if( tn->count > 0 ) {
    buf[d] = '\0';
    fprintf( fp, "%20d %s\n", tn->count, buf );
  }

  for( i = 0; i < NCHARS; ++i ) {
    buf[d] = (char)i;
    TriePreorder( tn->next[(unsigned char)i], buf, d + 1, fp );
  }

} /* TriePreorder */


/*============================================================*/
/* Save a trie to a file. Returns 0 on success, -1 on failure */
/*============================================================*/
int
TrieSave( TrieNode tn, char *filename ) {

  FILE *fp;
  char buf[BUFSIZ];

  /* open destination file */
  if( !(fp=fopen(filename, "w")) ) {
    errlogf( ERN, "TrieSave(): error writing trie to file %s\n", filename );
    return( -1 );
  }

  /* traverse the trie and write all the strings to fp */
  TriePreorder( tn, buf, 0, fp );

  /* close the destination file when done writing */
  fclose( fp );

  /* return success */
  return( 0 );

} /* TrieSave */


/*=====================================*/
/* Delete a trie (postorder traversal) */
/*=====================================*/
TrieNode
TrieDelete( TrieNode tn ) {

  int i;

  if( !tn )
    return( NULL );

  for( i = 0; i < NCHARS; ++i )
    TrieDelete( tn->next[(unsigned char)i] );

  free( tn );

  return( NULL );

} /* TrieDelete */

/*==================================================================*/
/* advance one char in the trie. Return the next node (can be NULL) */
/*==================================================================*/
TrieNode
TrieNextChar( TrieNode tn, char c ) {

  /* we're already in a "dead end" (i.e., not a valid word). Stay there */
  if( !tn )
    return( NULL );

  return( tn->next[(unsigned char)c] );

} /* TrieNextChar */


/*========================================*/
/* Return the occurence count of a string */
/*========================================*/
int
TrieStringCount( TrieNode tn ) {

  /* dead-end state -> string does not exist, count is 0 */
  if( !tn )
    return( 0 );

  return( tn->count );

} /* TrieStringCount */
