/* This file is a component of the SurfAgent Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*********************************  urlstr.c  **********************************

    Purpose: String operations on URL's.

    Notes:   none.
**/

/*includes*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "urlstr.h"
#include "miscstr.h"
#include "errlog.h"
#include "strlist.h"
#include "bindconn.h"

/*private defines (the public ones are in template.h)*/


/*private macros (the public ones are in template.h)*/


/*private typedefs (the public ones are in template.h)*/


/*public global variables (to be declared "external" in template.h)*/


/*private global variables (only for this file, declare as "static")*/


/*private functions (only for this file, declare as "static")*/

/*---------------------------------------------------------------------------*/
/* Simplifies the string describing a path. e.g., converts "/a/b/.././c"     */
/* into "/a/c" .                                                             */
/* Returns: char *: the simplified, converted path.                          */
/* Parameters: char *path - in - the original path;                          */
/* Notes: We assume the path starts with a '/', but we need to check whether */
/*        it ends with one, and append it if so...                           */
/*---------------------------------------------------------------------------*/
static char *
compress_path( char *path ) {

  char *buf = NULL;
  char *v[BUFSIZ];
  int i, n, ok, slash;
  StrList stack;

  if( !path ) {
    errlogf( ERR, "compress_path(): WARNING: NULL argument, returning NULL\n" );
    return( NULL );
  }

  if( !(stack = StrListCreate()) ) {
    errlogf( ERR, "compress_path(): failed to create stack...\n" );
    exit( 1 );
  }

  /* make a copy of the path */
  buf = strdup( path );

  /* check if we have a trailing '/' character */
  slash = (buf[strlen(buf) - 1] == '/');

  /* compress path components using the stack */
  n = ssplit( buf, "/", v, SZ(v), 1, 1 );
  for( i=0; i < n; ++i ) {
    ok = 1;
    if( strcmp(v[i], ".") == 0 )
      continue;
    else if( strcmp(v[i], "..") == 0 )
      ok = ( StrListPop(stack) == 0 );
    else
      ok = ( StrListAppend(stack, v[i]) == 0 );
    if( !ok ) {
      errlogf( ERR, "compress_path(): WARNING: invalid path: %s; "
                    "returning original path !!!\n",
                    path );
      StrListDestroy( stack );
      free( buf );
      return( strdup(path) );
    }
  }

  /* we're done with buf now */
  freez( buf );

  /* rebuild compressed path from stack */
  n = StrListSize( stack );
  for( i = 0; i < n; ++i ) {
    buf = strsav( buf, "/" );
    buf = strsav( buf, StrListPeek(stack, i) );
  }
  if( slash )
    buf = strsav( buf, "/" );

  /* destroy up the stack */
  StrListDestroy( stack );

  if( !buf ) {
    errlogf( ERR, "compress_path(): path=%s reduces to NULL -> returning /\n" );
    buf = strdup( "/" );
  }

  return( buf );

} /* compress_path */


/*public functions (to be declared "external" in template.h)*/

/*-----------------------------------------------------------------------*/
/* Extract the components of a http or https URL: host, port, and path   */
/* Returns: UrlContent - pointer to structure containing URL components  */
/* Paramenters: char *url - in - the url to parse;                       */
/* Notes: For now, only http and https are implemented. For any other    */
/*        protocol, NULL is returned for all other components.           */
/*-----------------------------------------------------------------------*/
UrlContent
parse_url( char *url ) {

  char *buf, *crtbuf;
  char *aux;
  UrlContent url_content;

  /* create url_content structure */
  if( !(url_content = malloc(sizeof(UrlContentStruct))) ) {
    errlogf( ERN, "parse_url(): malloc error...\n" );
    exit( 1 );
  }

  /* make a copy of the url, so we can analyze it non-destructively :) */
  buf = crtbuf = strdup( url );

  /* find the end of the protocol */
  if( !(aux = strstr(crtbuf, "://")) ) {
    errlogf( ERR, "parse_url(): can't get protocol in URL: %s\n", url );
    free( buf );
    free( url_content );
    return( NULL );
  }

  /* We now have the protocol */
  *aux = '\0';
  url_content->proto = strdup( crtbuf );

  if( (strcmp(crtbuf, "http") != 0) && (strcmp(crtbuf, "https") != 0) ) {
    errlogf( ERR, "parse_url(): unsupported protocol: %s\n", crtbuf );
    free( buf );
    free( url_content );
    return( NULL );
  }

  /* skip the "://" part */
  crtbuf = aux + 3;

  /* Now we're looking at the hostport/path part */

  /* get the path first */
  if( (aux = strchr(crtbuf, '/')) ) {
    url_content->path = compress_path( aux );
    *aux = '\0';
  } else {
    url_content->path = strdup( "/" );
  }

  /* get the hostport */
  url_content->hostport = strdup( crtbuf );

  /* now get the port */
  if( (aux = strchr(crtbuf, ':')) ) {
    url_content->port = atoi( aux + 1 );
    *aux = '\0';
  } else {
    url_content->port = 80;	/* or the default for whatever the protocol */
  }

  /* and finally the host itself */
  url_content->hoststr = strdup( crtbuf );

  if( (url_content->hostip = atoip(crtbuf)) == -1 ) {
    errlogf( ERN, "parse_url(): DNS lookup failed for host %s\n", crtbuf );
  }

  /* clean up buffer */
  free( buf );

  return( url_content );

} /* parse_url */


/*--------------------------------------------------------------------*/
/* Deletes a dynamically created UrlContent structure.                */
/* Returns NULL                                                       */
/* Parameters: UrlContent url_content - in - the structure to clear.  */
/* Notes: This function is designed to be used as follows:            */
/*           some_url_content = DeleteUrlContent( some_url_content ); */
/*--------------------------------------------------------------------*/
UrlContent
DeleteUrlContent( UrlContent url_content ) {

  if( url_content ) {

    freez( url_content->proto );
    freez( url_content->hostport );
    freez( url_content->hoststr );
    freez( url_content->path );

    free( url_content );

  }

  return( NULL );

} /* DeleteUrlContent */


/*----------------------------------------------------------------------------*/
/* Compares two URL structures to each other.                                 */
/* Returns: 0 if equal, non-zero if different.                                */
/* Parameters: UrlContent crtstruct, docstruct - in - the URLs to compare;    */
/* Notes: This routine tries to handle the presence or absence of a trailing  */
/*        '/' character in a halfway decent/intelligent manner. We assume     */
/*        that crturl is correct and has the '/' if the path is a directory.  */
/*        We then report equality if the docurl is the same except for the    */
/*        lack of the trailing slash. The assumption is that inside documents */
/*        people might omit the trailing '/', but the crturl comes from the   */
/*        server and is correct.                                              */
/*----------------------------------------------------------------------------*/
int
urlcmpstr( UrlContent crtstruct, UrlContent docstruct ) {

  int diff, len;

  /* compare protocols */
  /* Q: Should we consider http and https equal to each other for the purpose
   *    of this comparison ?
   */
  if( (diff = strcmp(crtstruct->proto, docstruct->proto)) != 0 )
    return( diff );

  /* compare hosts */
  if( (diff = crtstruct->hostip - docstruct->hostip) != 0 )
    return( diff );
  if( crtstruct->hostip == -1 ) {
    /* we need to strcasecmp the hoststr because of DNS lookup failure */
    if( (diff = strcasecmp(crtstruct->hoststr, docstruct->hoststr)) != 0 )
      return( diff );
  }

  /* compare ports */
  if( (diff = crtstruct->port - docstruct->port) != 0 )
    return( diff );

  /* compare paths */
  if( (diff = strcmp(crtstruct->path, docstruct->path)) == 0 )
    return( diff );		/* perfect match, we're done */

  /* now try without the trailing slash in crtstruct.path */
  if( ((len = strlen(docstruct->path)) == (strlen(crtstruct->path) - 1)) &&
      (crtstruct->path[len] == '/') &&
      (strncmp(crtstruct->path, docstruct->path, len) == 0) ) {
    return( 0 );
  }

  /* paths are clearly different, return diff from original path comparison...*/
  return( diff );

} /* urlcmpstr */


/*----------------------------------------------------------------------------*/
/* Compares two URL strings to each other.                                    */
/* Returns: 0 if equal, non-zero if different.                                */
/* Parameters: char *crturl, *docurl - in - the URLs to compare;              */
/* Notes: This routine uses urlcmpstr                                         */
/*----------------------------------------------------------------------------*/
int
urlcmp( char *crturl, char *docurl ) {

  UrlContent crtstruct, docstruct;
  int diff;

  /* parse both URLs */
  if( !(crtstruct = parse_url(crturl)) ||
      !(docstruct = parse_url(docurl)) ) {
    errlogf( ERR, "urlcmp(): failed to parse URL\n" );
    exit( 1 );
  }

  /* compare */
  diff = urlcmpstr( crtstruct, docstruct );

  /* clean up and return... */
  DeleteUrlContent( crtstruct );
  DeleteUrlContent( docstruct );
  return( diff );

} /* urlcmp */


/*--------------------------------------------------------------------------*/
/* Generates a fully qualified URL path out of a base and a relative path.  */
/* Returns: char * - a newly allocated string containing the result, or     */
/*                   NULL on error;                                         */
/* Parameters: char *basepath - in - the path of the base URL;              */
/*             char *relpath - in - the relative path;                      */
/* Notes: We assume the base is correct, (i.e., it has a '/' after the last */
/*        directory component of the path, and everything that follows the  */
/*        last '/' is the name of the file the base URL refers to...)       */
/*        Also, if the relpath starts with a '/' then it completely replaces*/
/*        the base path.                                                    */
/*--------------------------------------------------------------------------*/
char *
qualify_path( char *basepath, char *relpath ) {

  char *buffer, *aux;

  /* if the relpath starts with a slash, we'll just return it */
  if( *relpath == '/' )
    return( compress_path(relpath) );

  /* first, make a copy of the base path */
  buffer = strdup( basepath );

  /* Find the last occurrence of '/' in the base path and trim */
  if( !(aux = strrchr(buffer, '/')) ) {
    errlogf( ERR, "qualify_path(): invalid base path %s\n", buffer );
    return( NULL );
  }
  *(aux + 1) = '\0';

  /* concatenate the base and the rel */
  buffer = strsav( buffer, relpath );

  /* compress the resulting path (remove /./'s, and compress /../'s) */
  aux = compress_path( buffer );

  /* de-allocate junk */
  free( buffer );

  /* return the final version of the fully qualified path */
  return( aux );

} /* qualify_path */


/*-------------------------------------------------------------------------*/
/* This function tries to figure out if url is a fully qualified one, or   */
/* just a relative path for another base url.                              */
/* Returns: 1 (True) if fully qualified, 0 if relative.                    */
/* Paramenters: char *url - in - the url we want to find out about...      */
/* Notes: Right now, only http and https are scanned for. In the future we */
/*        might want to look at other protocols, such as ftp, mailto, etc. */
/*-------------------------------------------------------------------------*/
int
is_qualified( char *url ) {

  if( (strncmp("http://", url, 7) == 0) ||
      (strncmp("https://", url, 8) == 0) )
    return( 1 );

  return( 0 );

} /* is_qualified */


/*--------------------------------------------------------------------------*/
/* Generates a fully qualified URL structure using a base structure and a   */
/* relative path spec.                                                      */
/* Returns: char * - a newly allocated string containing the result, or     */
/*                   NULL on error;                                         */
/* Parameters: UrlContent base - in - the parsed structure of the base URL  */
/*             char *relpath - in - the relative path;                      */
/* Notes: We assume the base is correct, (i.e., it has a '/' after the last */
/*        directory component of the path, and everything that follows the  */
/*        last '/' is the name of the file the base URL refers to...)       */
/*--------------------------------------------------------------------------*/
UrlContent
qualify_url( UrlContent base, char *relpath ) {

  UrlContent result;

  /* Create result structure */
  if( !(result = malloc(sizeof(UrlContentStruct))) ) {
    errlogf( ERN, "qualify_url(): malloc error...\n" );
    exit( 1 );
  }

  /* Copy the relevant fields from the base */
  result->proto    = strdup( base->proto );
  result->hostport = strdup( base->hostport );
  result->hoststr  = strdup( base->hoststr );
  result->hostip   = base->hostip;
  result->port     = base->port;

  /* the path will be the result of qualify_path() */
  result->path = qualify_path( base->path, relpath );

  /* return newly created URL structure */
  return( result );

} /* qualify_url */
