/* This file is a component of the SurfAgent Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/********************************** wordvec.c **********************************

    Purpose: Word vector abstract data type implementation.

    Notes:   This module implements a word-vector with word counts. It is
             intended for frequent insertions and updates, and therefore
             the underlying structure is a list sorted alphabetically.
**/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "wordvec.h"
#include "errlog.h"


/*******************************************************************************
    WordListInsCnt( v, word, count )

    Purpose:   add word to list v, count times. If the word is not contained
               in v yet, it will be added and its count set appropriately.
               otherwise, the count will be incremented
               The list is maintained alphabetically sorted
    Returns:   WordList - a pointer to the updated word list (NULL on error);
    Arguments: WordList v   - in: word list to which addition is requested;
               char *word   - in: the word we want to add/update in the list;
               double count - in: set or increment the occurrence count of the
                                  word as specified by flag;
**/
static WordList
WordListInsCnt( WordList v, char *word, double count ) {

  WordList vnew = NULL, p, q;
  int cmp = -1, action = 0;

  /* verify parameter sanity */
  if( !word || (*word == '\0') ) {
    errlogf( ERR, "WordListInsCnt(): warning: NULL ptr or empty string "
                  "insertion ignored.\n" );
    return( v );
  }

  /* find the first word in _v_ greater than or equal to _word_ */
  for( p = v, q = NULL;
       p && (cmp=strcmp(p->word, word))<0;
       q = p, p = p->next );

  /* figure out the action given p, q, and cmp */
  if( p )   action += 4;
  if( q )   action += 2;
  if( cmp ) action += 1;

  /* create new node if needed (according to action) */
  if( (0<=action && action<=3) || action==5 || action==7 ) {
    if( !(vnew=malloc(sizeof(WordListStruct))) ) {
      errlogf( ERN, "WordListInsCnt(): malloc error\n" );
      exit( 1 );
    }
    vnew->word = strdup( word );
    vnew->count = count;		/* if this is a newly entered word, its
					 * previous count is 0, so it doesn't
					 * matter if we increment or set !
					 */
    vnew->next = NULL;
  }

  /* perform the operations indicated by action */
  switch( action ) {
    case 0:			/* insert at the front (or create if p==NULL) */
    case 1:
    case 5:
      vnew->next = p;
      break;
    case 2:			/* ins. betw. q and p (at the end if p==NULL)*/
    case 3:
    case 7:
      vnew->next = p;
      q->next = vnew;
      vnew = v;			/* return original v */
      break;
    case 4:			/* found word at pointer p - inc. p->count */
    case 6:
      p->count += count;
      vnew = v;			/* return original v */
      break;
    default:			/* should never get here */
      errlogf( ERR, "WordListInsCnt(): error adding to WordVec ...\n" );
      exit( 1 );		/* return error */
      break;
  }

  /* return updated word list */
  return( vnew );

} /* WordListInsCnt */


/*******************************************************************************
    WordVecNew( )

    Purpose:   Create a new word vector.
    Returns:   WordVec        - a newly created word vector, or NULL if error;
    Arguments: none;
**/
WordVec
WordVecNew( void ) {

  WordVec wv;

  if( !(wv = malloc(sizeof(WordVecStruct))) ) {
    errlogf( ERN, "WordVecNew(): error allocating new word vector!\n" );
    return( NULL );
  }

  wv->wordlist = NULL;
  wv->length = 0.0;

  return( wv );

} /* WordVecNew */


/*******************************************************************************
    WordVecLoad( filename )

    Purpose:   Load a word vector from a file.
    Returns:   WordVec        - a newly created word vector, loaded from a file,
                                or NULL if an error occurs.
    Arguments: char* filename - in: the name of a file from which to load;
**/
WordVec
WordVecLoad( char *filename ) {

  WordVec wv;
  WordList last_v = NULL, new_v;
  FILE *fp;
  char word[BUFSIZ + 1];
  double count, sum = 0.0;

  /* create empty word vector */
  if( !(wv = WordVecNew()) ) {
    errlogf( ERN, "WordVecLoad(): error creating new word vector.\n" );
    return( NULL );
  }

  /* open source file */
  if( !(fp=fopen(filename, "r")) ) {
    errlogf( ERN, "WordVecLoad(): error opening word vector file %s\n",
                  filename );
    return( wv );
  }

  /* read the file: each line is a pair of count and word */
  while( fscanf(fp, "%lf%s", &count, word) == 2 ) {

    word[BUFSIZ] = '\0';	/* make sure string is terminated */

    /* create new node */
    if( !(new_v = malloc(sizeof(WordListStruct))) ) {
      errlogf( ERN, "WordVecLoad(): malloc() error\n" );
      exit( 1 );	/* the nice way should be return( NULL ); */
    }
    new_v->word = strdup( word );
    new_v->count = count;
    new_v->next = NULL;

    sum += count * count;

    /* insert at the end, verify if in alphabetical order */
    if( wv->wordlist ) {
      if( strcmp(last_v->word, word) >= 0 ) {
        errlogf( ERR, "WordVecLoad(): WordVec file '%s' not in alphabetical "
                      "order (%s - %s)\n",
                      filename, last_v->word, word );
        exit( 1 );
      }
      last_v->next = new_v;
    } else {
      wv->wordlist = new_v;
    }
    last_v = new_v;

  }

  /* close the input file when done */
  fclose( fp );

  wv->length = sqrt( sum );

  /* return the resulting vector */
  return( wv );

} /* WordVecLoad */


/*******************************************************************************
    WordVecSave( wv, filename )

    Purpose:   Save word vector v to a file indicated by its filename.
    Returns:   int            - 0 on success, -1 on failure (FIXED!!!)
    Arguments: WordVec wv     - in: the vector to be saved
               char* filename - in: the name of the file where we wish to save;
    Notes: Saving doubles with fprintf loses precision, but is portable. If this
           were a single-platform implementation, it would make more sense to
           use write and sizeof(double)...
**/
int
WordVecSave( WordVec wv, char *filename ) {

  WordList p;
  FILE *fp;

  /* sanity check */
  if( !wv ) {
    errlogf( ERR, "WordVecSave(): can't save null WordVec !\n" );
    return( -1 );
  }

  /* open destination file */
  if( !(fp=fopen(filename, "w")) ) {
    errlogf( ERN, "WordVecSave(): error writing word vector to file %s\n",
                  filename );
    return( -1 );
  }

  /* write all words and their counts to the file */
  for( p = wv->wordlist; p; p = p->next )
    fprintf( fp, "%41.20f %s\n", p->count, p->word );

  /* close the destination file when done writing */
  fclose( fp );

  /* return success */
  return( 0 );

} /* WordVecSave */


/*******************************************************************************
    WordVecDelete( wv )

    Purpose:   Free the memory used by a word vector.
    Returns:   WordVec    - NULL pointer, so that we can use this function like
                            this:  v = WordVecDelete( v ); - and wv would end up
                            being NULL;
    Arguments: WordVec wv - in: the vector to be deleted
**/
WordVec
WordVecDelete( WordVec wv ) {

  WordList p, q;

  if( wv ) {

    /* deallocate each word in the list */
    p = wv->wordlist;
    while( p ) {
      q = p;
      p = p->next;
      free( q->word );
      free( q );
    }

    free( wv );
  }

  /* return NULL (the new value for vector v) */
  return( NULL );

} /* WordVecDelete */


/*******************************************************************************
    WordVecInsert( wv, word )

    Purpose:   add word to vector wv. If the word is not contained
               in wv yet, it will be added and its count set to 1.
               otherwise, the count will be incremented by 1.
    Returns:   int - 0 on success, -1 on error;
    Arguments: WordVec wv - in: word vector to which addition is requested;
               char *word - in: the word we want to add/update in the vector;
**/
int
WordVecInsert( WordVec wv, char *word ) {

  /* sanity check */
  if( !wv ) {
    errlogf( ERR, "WordVecInsert(): Can't insert into NULL WordVec !\n" );
    return( -1 );
  }

  wv->wordlist = WordListInsCnt( wv->wordlist, word, 1.0 );

  return( 0 );

} /* WordVecInsert */


/*******************************************************************************
    WordVecSumWtNrm( wv1, wt1, nrm1, wv2, wt2, nrm2 )

    Purpose:   Add together two word vectors. The result should be a word vector
               containing the union of words from wv1 and wv2, with counts equal
               to the sum between the counts in wv1 and those in wv2 (if a word
               is not present in a vector, its count is considered 0).
               Before addition, the two operand vectors can be optionally
               weighted and normalized.
               The length of the sum vector will be recomputed from the counts
               of the components.
    Returns:   WordVec        - a newly created word vector, the sum/union of
                                wv1 and wv2 (a new vector is created, to
                                preserve wv1 and wv2 intact for further use);
                                in case of any error, NULL is returned;
    Arguments: WordVec wv1, wv2 - in: the two (constant) vectors to be added;
               double wt1, wt2  - in: weights to be applied to the vectors;
               int nrm1, nrm2   - in: boolean flags for normalization;
**/
WordVec
WordVecSumWtNrm( WordVec wv1, double wt1, int nrm1,
                 WordVec wv2, double wt2, int nrm2 ) {

  WordVec wvsum;
  WordList p1, p2, new_v, last_v = NULL ;  /* <-NULL avoids compiler warning */
  double dim1, dim2, len, sum = 0.0;
  int cmp;

  /* sanity check */
  if( !(wv1 && wv2) ) {
    errlogf( ERR, "WordVecSumWtNrm(): invalid operands: %p or %p\n", wv1, wv2 );
    return( NULL );
  }

   /* create empty word vector */
  if( !(wvsum = WordVecNew()) ) {
    errlogf( ERN, "WordVecSumWtNrm(): error creating new word vector.\n" );
    return( NULL );
  }

  /* add common parts of the two vectors */
  for( p1 = wv1->wordlist, p2 = wv2->wordlist; p1 && p2; ) {

    /* create new node */
    if( !(new_v = malloc(sizeof(WordListStruct))) ) {
      errlogf( ERN, "WordVecSumWtNrm(): malloc() error\n" );
      exit( 1 );	/* the nice way should be return( NULL ); */
    }
    new_v->next = NULL;

    cmp = strcmp( p1->word, p2->word );

    if( cmp < 0 ) {			/* p1<p2 -> advance p1 */
      new_v->word = strdup( p1->word );
      new_v->count = p1->count * wt1;
      if( nrm1 )
        new_v->count /= wv1->length;
      p1 = p1->next;
    }
    if( cmp > 0 ) {			/* p2<p1 -> advance p2 */
      new_v->word = strdup( p2->word );
      new_v->count = p2->count *wt2;
      if( nrm2 )
        new_v->count /= wv2->length;
      p2 = p2->next;
    }
    if( cmp == 0 ) {			/* p1=p2 -> advance both */
      new_v->word = strdup( p1->word );
      dim1 = p1->count * wt1;
      if( nrm1 )
        dim1 /= wv1->length;
      dim2 = p2->count * wt2;
      if( nrm2 )
        dim2 /= wv2->length;
      new_v->count = dim1 + dim2;
      p1 = p1->next;
      p2 = p2->next;
    }

    sum += new_v->count * new_v->count;

    /* insert at the end */
    if( wvsum->wordlist ) {
      last_v->next = new_v;
    } else {
      wvsum->wordlist = new_v;
    }
    last_v = new_v;

  } /* done with common parts of the vectors */

  /* copy the rest of whichever list remains */
  len = wv1->length;
  if( !p1 ) {
    p1 = p2;
    wt1 = wt2;
    nrm1 = nrm2;
    len = wv2->length;
  }

  for( ; p1; p1 = p1->next ) {

    /* create new node */
    if( !(new_v = malloc(sizeof(WordListStruct))) ) {
      errlogf( ERN, "WordVecSumWtNrm(): malloc() error\n" );
      exit( 1 );	/* the nice way should be return( NULL ); */
    }
    new_v->next = NULL;

    new_v->word = strdup( p1->word );
    new_v->count = p1->count * wt1;
    if( nrm1 )
      new_v->count /= len;

    sum += new_v->count * new_v->count;

    /* insert at the end */
    if( wvsum->wordlist ) {
      last_v->next = new_v;
    } else {
      wvsum->wordlist = new_v;
    }
    last_v = new_v;
  }

  wvsum->length = sqrt( sum );

  /* return the result */
  return( wvsum );

} /* WordVecSumWtNrm */


/*******************************************************************************
    WordVecSimilarity( wv1, wv2 )

    Purpose:   Compute the similarity metric ( double no. in [0,1] ) of two
               word vectors (we assume they are tf-idf vectors). This is
               actually accomplished by computing the dot-product between them.
    Returns:   double - the similarity metric between the two vectors. This is
                        a double precision real number from 0.0 (not similar) to
                        1.0 (similar);
    Arguments: WordVec wv1, wv2 - in: the two (constant) vectors to be compared;
**/
double
WordVecSimilarity( WordVec wv1, WordVec wv2 ) {

  WordList p1, p2;
  double lenprod, sum12 = 0.0;
  int cmp;

  /* sanity check */
  if( !(wv1 && wv2) ) {
    errlogf( ERR, "WordVecSimilarity(): invalid operands: %p or %p\n",
                  wv1, wv2 );
    exit( 1 );
  }
 
  p1 = wv1->wordlist;
  p2 = wv2->wordlist;

  /* if any of the vectors is empty, similarity is 0 */
  if( !(p1 && p2) )
    return( 0.0 );

  while( p1 && p2 ) {

    /* compare p1 and p2 */
    cmp = strcmp( p1->word, p2->word );

    if( cmp < 0 ) {			/* p1<p2 -> advance p1 */
      p1 = p1->next;
    }

    if( cmp > 0 ) {			/* p2<p1-> advance p2 */
      p2 = p2->next;
    }

    if( cmp == 0 ) {			/* p1=p2 -> advance both */
      sum12 += ( p1->count * p2->count );
      p1 = p1->next;
      p2 = p2->next;
    }

  }

  /* more sanity check */
  if( (lenprod = wv1->length * wv2->length) == 0.0 ) {
    errlogf( ERR, "WordVecSimilarity(): impossible wv length:"
                  " %20.10f, %20.10f\n", wv1->length, wv2->length );
    exit( 1 );
  }

  /* compute and return similarity metric */
  return( sum12 / (wv1->length * wv2->length) );

} /* WordVecSimilarity */
